"""Calculates intramolecular forces between bonded particles in molecules
"""
import numpy as np
import networkx as nx
from dataclasses import dataclass

# Imported here so we can call from force import compute_bond_forces__fortran
from force_kernels import cbf as compute_bond_forces__fortran  # noqa: F401
from force_kernels import (  # noqa: F401
    caf as compute_angle_forces__fortran,
)
from force_kernels import (  # noqa: F401
    cdf as compute_dihedral_forces__fortran,
)
from force_kernels import (  # noqa: F401
    cbf_d as compute_bond_forces__fortran__double,
)
from force_kernels import (  # noqa: F401
    caf_d as compute_angle_forces__fortran__double,
)
from force_kernels import (  # noqa: F401
    cdf_d as compute_dihedral_forces__fortran__double,
)


@dataclass  # might not be needed
class Dielectric_type:
    atom_1: str
    dielectric_value: float


@dataclass
class Bond:
    """Dataclass representing a single two-particle bond type

    A `bond type` is a bond strength and equilibrium distance associated with
    any bond between particles of specific types :code:`A` and :code:`B`
    (where :code:`A` and :code:`B` may be the same or different). Harmonic
    two-particle bonds in HyMD take the form

    .. math::

        V_2(\\mathbf{r}_1, \\mathbf{r}_2) =
            \\frac{1}{2}k
            \\left(
                \\vert \\mathbf{r}_1 - \\mathbf{r}_2 \\vert - r_0
            \\right)^2,

    where :math:`k` is the bond strength (spring constant) and :math:`r_0` is
    the equilibrium bond length (at which the energy is zero).

    Attributes
    ----------
    atom_1 : str
        Type name of particle 1.
    atom_2 : str
        Type name of particle 2.
    equilibrium : float
        Equilibrium distance at which the energy associated with the bond
        vanishes and the resulting force is zero.
    strength : float
        Harmonic bond strength coefficient (spring constant).
    """

    atom_1: str
    atom_2: str
    equilibrium: float
    strength: float


@dataclass
class Angle(Bond):
    """Dataclass representing a single three-particle bond type

    A `bond type` is a bond strength and equilibrium angle associated with
    any three-particle bond between particles of specific types :code:`A`,
    :code:`B`, and :code:`C` (where :code:`A`, :code:`B`, and :code:`C` may be
    different or the same). Harmonic angular three-particle bonds in HyMD take
    the form

    .. math::

        V_3(\\mathbf{r}_1, \\mathbf{r}_2, \\mathbf{r}_3) =
            \\frac{1}{2}k
            \\left(
                \\cos^{-1}
                \\left[
                    \\frac{
                        (\\mathbf{r}_1-\\mathbf{r}_2)
                        \\cdot
                        (\\mathbf{r}_3-\\mathbf{r}_2)
                    }
                    {
                        \\vert\\mathbf{r}_1-\\mathbf{r}_2\\vert
                        \\vert\\mathbf{r}_3-\\mathbf{r}_2\\vert
                    }
                \\right] - \\theta_0
            \\right)^2

    Attributes
    ----------
    atom_1 : str
        Type name of particle 1.
    atom_2 : str
        Type name of particle 2.
    atom_3 : str
        Type name of particle 3.
    equilibrium : float
        Equilibrium angle at which the energy associated with the
        three-particle angular bond vanishes and the resulting force is zero.
    strength : float
        Harmonic bond strength coefficient.

    See also
    --------
    Bond :
        Two-particle bond type dataclass
    """

    atom_3: str


@dataclass
class Dihedral:
    """Dataclass representing a single four-particle bond type

    A `bond type` is a bond strength and equilibrium angle associated with
    any four-particle torsional bond between particles of specific types
    :code:`A`, :code:`B`, :code:`C`, and :code:`D` (where :code:`A`, :code:`B`,
    :code:`C`, and :code:`D` may be different or the same). Dihedral
    four-particle bonds in HyMD take different forms depending on the
    :code:`dih_type` parameter.

    In the following, let :math:`\\phi` denote the angle between the planes
    spanned by the relative positions of atoms :code:`A`-:code:`B`-:code:`C`
    and :code:`B`-:code:`C`-:code:`D`. If :code:`dih_type = 0`, then

    .. math::

        V_4(\\phi) = \\sum_n c_n
            \\left(
                1 + \\cos\\left[
                    n\\phi - d_n
                \\right]
            \\right),

    where :math:`c_n` are energy coefficients and :math:`d_n` are propensity
    phases. By default, the cosine sum is truncated at five terms. If
    :code:`coeffs` provides only a single float, this is used as the coiling
    propensity parameter :math:`\\lambda`. In this case, :math:`c_n` and
    :math:`d_n` are automatically set to values which promote alpha helical
    (:math:`\\lambda=-1`), beta sheet (:math:`\\lambda=1`), or a mixed
    (:math:`1>\\lambda>-1`) structure (using values provided by Bore et al.
    (2018)). In this case, the full potential takes the form

    .. math::

        V_4(\\phi;\\lambda) =
            \\frac{1}{2}(1-\\lambda) V_{\\text{prop},\\alpha}(\\phi)
            +
            \\frac{1}{2}(1+\\lambda) V_{\\text{prop},\\beta}(\\phi)
            +
            (1-\\vert\\lambda\\vert) V_{\\text{prop}, \\text{coil}}(\\phi),

    with each of the :math:`V_{\\mathrm{prop}, X}` being fixed cosine series
    potentials with pre-set :math:`c_n` -s and :math:`d_n` -s.

    If :code:`dih_type = 1`, then a combined bending torsional (CBT) potential
    is employed,

    .. math::

        V_4(\\phi,\\gamma;\\lambda) =
            V_4(\\phi;\\lambda)
            +
            \\frac{1}{2}K(\\phi)(\\gamma - \\gamma_0)^2,

    where :math:`V_4(\\phi;\\lambda)` specifies the potential as given by the
    coiling propensity parameter above, :math:`K(\\phi)` is a separate cosine
    series potential acting as the angular three-particle bond strength, and
    :math:`\\gamma_0` is the three-particle bond equilibrium angle. In this
    case, the :code:`coeffs` parameter must specify *both* a :math:`\\lambda`
    value, in additon to the energy and phases dictating the :math:`K(\\phi)`
    potential.

    Attributes
    ----------
    atom_1 : str
        Type name of particle 1.
    atom_2 : str
        Type name of particle 2.
    atom_3 : str
        Type name of particle 3.
    atom_4 : str
        Type name of particle 4.
    coeffs : list[list[float]] or list[float] or numpy.ndarray
        Dihedral coefficients defining the series expansion of the dihedral
        energy.
    dih_type : int
        Specifies the type of dihedral used; If :code:`0`, then :code:`coeffs`
        must contain **either** two lists of five energy coefficients
        :math:`c_n` and five propensity phases :math:`d_n` **or** a single
        floating point number defining the :math:`\\lambda` coiling propensity
        parameter. If :code:`1`, the combined bending-torsional potential is
        used, and :code:`coeffs` must specify :math:`\\lambda` *and* two lists
        containing energy coefficients and propensity phases for the
        :math:`V_\\text{prop}` propensity potential, defined in terms of cosine
        series.

    References
    ----------
    Bore et al. J. Chem. Theory Comput., 14(2): 1120–1130, 2018.
    """

    atom_1: str
    atom_2: str
    atom_3: str
    atom_4: str
    coeffs: np.ndarray
    dih_type: int


@dataclass
class Chi:
    """Dataclass representing a single :math:`\\chi` mixing interaction type

    An `interaction mixing energy type` is a mixing energy associated with
    density overlap between species of types :code:`A` and :code:`B`
    (specified as inputs :code:`atom_1` and :code:`atom_2`). A positive mixing
    energy promotes phase separation, a negative mixing energy promotes mixing.
    The interaction energy density (provided the :code:`DefaultWithChi`
    Hamiltonian is used) takes the form

    .. math::

        w[\\tilde\\phi] = \\frac{1}{2\\kappa}
            \\sum_{k,l}\\chi_{k,l} \\tilde\\phi_k\\tilde\\phi_l,

    where :math:`\\chi_{k,l}` denotes the mixing energy between species
    :math:`k` and :math:`l`, with :math:`\\kappa` being the incompressibility.
    The value of the interaction mixing energy parameter may be extracted from
    `Flory-Huggins theory`_.

    .. _`Flory-Huggins theory`:
        https://en.wikipedia.org/wiki/Flory%E2%80%93Huggins_solution_theory

    Attributes
    ----------
    atom_1 : str
        Type name of particle 1.
    atom_2 : str
        Type name of particle 2.
    interaction_energy : float
        Interaction mixing energy.

    See also
    --------
    hymd.hamiltonian.DefaultWithChi :
        Interaction energy functional using :math:`\\chi`-interactions.
    """

    atom_1: str
    atom_2: str
    interaction_energy: float


def find_all_paths(G, u, n):
    """Helper function that recursively finds all paths of given lenght 'n + 1' inside a network 'G'.
    Adapted from https://stackoverflow.com/a/28103735."""
    if n == 0:
        return [[u]]
    paths = []
    for neighbor in G.neighbors(u):
        for path in find_all_paths(G, neighbor, n - 1):
            if u not in path:
                paths.append([u] + path)
    return paths


def prepare_bonds_old(molecules, names, bonds, indices, config):
    """Find bonded interactions from connectivity and bond types information

    .. deprecated:: 1.0.0
        :code:`prepare_bonds_old` was replaced by :code:`prepare_bonds` for
        use with compiled Fortran kernels prior to 1.0.0 release.

    Prepares the necessary equilibrium and bond strength information needed by
    the intramolecular interaction functions. This is performed locally on each
    MPI rank, as the domain decomposition ensures that for all molecules *all*
    consituent particles are always contained on *the same* MPI rankself.

    This function traverses the bond connectivity information provided in the
    structure/topology input file and indentifies any two-, three-, or
    four-particle potential bonds. For each connected chain of two, three, or
    four particles, a matching to bond types is attempted. If the corresponding
    names match, a bond object is initialized.

    In order to investigate the connectivity, a graph is created using
    networkx functionality.

    Parameters
    ----------
    molecules : (N,) numpy.ndarray
        Array of integer molecule affiliation for each of :code:`N` particles.
        Global (across all MPI ranks) or local (local indices on this MPI rank
        only) may be used, both, without affecting the result.
    names : (N,) numpy.ndarray
        Array of type names for each of :code:`N` particles.
    bonds : (N,M) numpy.ndarray
        Array of :code:`M` bonds originating from each of :code:`N` particles.
    indices : (N,) numpy.ndarray
        Array of integer indices for each of :code:`N` particles. Global
        (across all MPI ranks) or local (local indices on this MPI rank only)
        may be used, both, without affecting the result.
    config : Config
        Configuration object.

    Returns
    -------
    bonds_2 : list
        List of lists containing *local* particle indices, equilibrium
        distance, and bond strength coefficient for each reconstructed
        two-particle bond.
    bonds_3 :
        List of lists containing *local* particle indices, equilibrium angle,
        and bond strength coefficient for each reconstructed three-particle
        bond.
    bonds_4 :
        List of lists containing *local* particle indices, dihedral type index,
        and dihedral coefficients for each reconstructed four-particle
        torsional bond.
    bb_index : list
        List indicating the dihedral type of each four-particle bond in
        :code:`bonds_4`.

    See also
    --------
    Bond :
        Two-particle bond type dataclass.
    Angle :
        Three-particle angular bond type dataclass.
    Dihedral :
        Four-particle torsional bond type dataclass.
    hymd.input_parser.Config
        Configuration dataclass handler.
    """
    bonds_2 = []
    bonds_3 = []
    bonds_4 = []
    bb_index = []

    different_molecules = np.unique(molecules)
    for mol in different_molecules:
        bb_dihedral = 0
        bond_graph = nx.Graph()

        for local_index, global_index in enumerate(indices):
            if molecules[local_index] != mol:
                continue

            bond_graph.add_node(
                global_index,
                name=names[local_index].decode("UTF-8"),
                local_index=local_index,
            )
            for bond in [b for b in bonds[local_index] if b != -1]:
                bond_graph.add_edge(global_index, bond)

        connectivity = nx.all_pairs_shortest_path(bond_graph)
        for c in connectivity:
            i = c[0]
            connected = c[1]
            for j, path in connected.items():
                if len(path) == 2 and path[-1] > path[0]:
                    name_i = bond_graph.nodes()[i]["name"]
                    name_j = bond_graph.nodes()[j]["name"]

                    for b in config.bonds:
                        match_forward = name_i == b.atom_1 and name_j == b.atom_2
                        match_backward = name_i == b.atom_2 and name_j == b.atom_1
                        if match_forward or match_backward:
                            bonds_2.append(
                                [
                                    bond_graph.nodes()[i]["local_index"],
                                    bond_graph.nodes()[j]["local_index"],
                                    b.equilibrium,
                                    b.strength,
                                ]
                            )

                if len(path) == 3 and path[-1] > path[0]:
                    name_i = bond_graph.nodes()[i]["name"]
                    name_mid = bond_graph.nodes()[path[1]]["name"]
                    name_j = bond_graph.nodes()[j]["name"]

                    for a in config.angle_bonds:
                        match_forward = (
                            name_i == a.atom_1
                            and name_mid == a.atom_2
                            and name_j == a.atom_3
                        )
                        match_backward = (
                            name_i == a.atom_3
                            and name_mid == a.atom_2
                            and name_j == a.atom_1
                        )
                        if match_forward or match_backward:
                            bonds_3.append(
                                [
                                    bond_graph.nodes()[i]["local_index"],
                                    bond_graph.nodes()[path[1]]["local_index"],
                                    bond_graph.nodes()[j]["local_index"],
                                    np.radians(a.equilibrium),
                                    a.strength,
                                ]
                            )

            all_paths_len_four = find_all_paths(bond_graph, i, 3)
            for p in all_paths_len_four:
                name_i = bond_graph.nodes()[i]["name"]
                name_mid_1 = bond_graph.nodes()[p[1]]["name"]
                name_mid_2 = bond_graph.nodes()[p[2]]["name"]
                name_j = bond_graph.nodes()[p[3]]["name"]

                for a in config.dihedrals:
                    match_forward = (
                        name_i == a.atom_1
                        and name_mid_1 == a.atom_2
                        and name_mid_2 == a.atom_3
                        and name_j == a.atom_4
                    )
                    if (
                        match_forward
                        and [
                            bond_graph.nodes()[p[3]]["local_index"],
                            bond_graph.nodes()[p[2]]["local_index"],
                            bond_graph.nodes()[p[1]]["local_index"],
                            bond_graph.nodes()[i]["local_index"],
                            a.coeffs,
                            a.dih_type,
                        ]
                        not in bonds_4
                    ):
                        bonds_4.append(
                            [
                                bond_graph.nodes()[i]["local_index"],
                                bond_graph.nodes()[p[1]]["local_index"],
                                bond_graph.nodes()[p[2]]["local_index"],
                                bond_graph.nodes()[p[3]]["local_index"],
                                a.coeffs,
                                a.dih_type,
                            ]
                        )
                        if a.dih_type == 1:
                            bb_dihedral = len(bonds_4)

        if bb_dihedral:
            bb_index.append(bb_dihedral - 1)

    return bonds_2, bonds_3, bonds_4, bb_index


def prepare_index_based_bonds(molecules, topol):
    bonds_2 = []
    bonds_3 = []
    bonds_4 = []

    different_molecules = np.unique(molecules)
    for mol in different_molecules:
        resid = mol + 1
        top_summary = topol["system"]["molecules"]
        resname = None
        test_mol_number = 0
        for molname in top_summary:
            test_mol_number += molname[1]
            if resid <= test_mol_number:
                resname = molname[0]
                break

        if resname is None:
            break

        if "bonds" in topol[resname]:
            first_id = np.where(molecules == mol)[0][0]
            for bond in topol[resname]["bonds"]:
                index_i = bond[0] - 1 + first_id
                index_j = bond[1] - 1 + first_id
                # bond[2] is the bond type, inherited by the itp format. Not used
                equilibrium = bond[3]
                strength = bond[4]
                bonds_2.append([index_i, index_j, equilibrium, strength])

        if "angles" in topol[resname]:
            first_id = np.where(molecules == mol)[0][0]
            for angle in topol[resname]["angles"]:
                index_i = angle[0] - 1 + first_id
                index_j = angle[1] - 1 + first_id
                index_k = angle[2] - 1 + first_id
                # angle[3] is the angle type, inherited by the itp format. Not used
                equilibrium = np.radians(angle[4])
                strength = angle[5]
                bonds_3.append([index_i, index_j, index_k, equilibrium, strength])

        if "dihedrals" in topol[resname]:
            first_id = np.where(molecules == mol)[0][0]
            for angle in topol[resname]["dihedrals"]:
                index_i = angle[0] - 1 + first_id
                index_j = angle[1] - 1 + first_id
                index_k = angle[2] - 1 + first_id
                index_l = angle[3] - 1 + first_id
                dih_type = angle[4]
                coeff = angle[5]
                bonds_4.append([index_i, index_j, index_k, index_l, coeff, dih_type, 0])
    return bonds_2, bonds_3, bonds_4


def prepare_bonds(molecules, names, bonds, indices, config, topol=None):
    """Rearrange the bond information for usage in compiled Fortran kernels

    Restructures the lists resulting from the execution of
    :code:`prepare_bonds_old` into numpy arrays suitable for calls to optimized
    Fortran code calculating bonded forces and energiesself.

    Parameters
    ----------
    molecules : (N,) numpy.ndarray
        Array of integer molecule affiliation for each of :code:`N` particles.
        Global (across all MPI ranks) or local (local indices on this MPI rank
        only) may be used, both, without affecting the result.
    names : (N,) numpy.ndarray
        Array of type names for each of :code:`N` particles.
    bonds : (N,M) numpy.ndarray
        Array of :code:`M` bonds originating from each of :code:`N` particles.
    indices : (N,) numpy.ndarray
        Array of integer indices for each of :code:`N` particles. Global
        (across all MPI ranks) or local (local indices on this MPI rank only)
        may be used, both, without affecting the result.
    config : Config
        Configuration object.

    Returns
    -------
    bonds_2_atom_1 : (B,) numpy.ndarray
        Local index of particle 1 for each of :code:`B` constructed
        two-particle bonds.
    bonds_2_atom_2 : (B,) numpy.ndarray
        Local index of particle 2 for each of :code:`B` constructed
        two-particle bonds.
    bonds_2_equilibrium : (B,) numpy.ndarray
        Equilibrium distance for each of :code:`B` constructed two-particle
        bonds.
    bonds_2_strength : (B,) numpy.ndarray
        Bond strength for each of :code:`B` constructed two-particle
        bonds.
    bonds_3_atom_1 : (A,) numpy.ndarray
        Local index of particle 1 for each of :code:`A` constructed
        three-particle bonds.
    bonds_3_atom_2 : (A,) numpy.ndarray
        Local index of particle 2 for each of :code:`A` constructed
        three-particle bonds.
    bonds_3_atom_3 : (A,) numpy.ndarray
        Local index of particle 3 for each of :code:`A` constructed
        three-particle bonds.
    bonds_3_equilibrium : (A,) numpy.ndarray
        Equilibrium angle for each of :code:`A` constructed three-particle
        bonds.
    bonds_3_strength : (A,) numpy.ndarray
        Bond strength for each of :code:`A` constructed three-particle
        bonds.
    bonds_4_atom_1 : (D,) numpy.ndarray
        Local index of particle 1 for each of :code:`D` constructed
        four-particle bonds.
    bonds_4_atom_2 : (D,) numpy.ndarray
        Local index of particle 2 for each of :code:`D` constructed
        four-particle bonds.
    bonds_4_atom_3 : (D,) numpy.ndarray
        Local index of particle 3 for each of :code:`D` constructed
        four-particle bonds.
    bonds_4_atom_4 : (D,) numpy.ndarray
        Local index of particle 4 for each of :code:`D` constructed
        four-particle bonds.
    bonds_4_coeff : (D,) numpy.ndarray
        Cosine series coefficients for each of :code:`D` constructed
        four-particle bonds.
    bonds_4_type : (D,) numpy.ndarray
        Dihedral type for each of :code:`D` constructed four-particle
        bonds.
    bonds_4_last : (D,) numpy.ndarray
        Flags indicating if :code:`dih_type` is :code:`1` for each of :code:`D`
        constructed four-particle bonds.

    See also
    --------
    prepare_bonds_old :
        Used internally to reconstruct the bonded interactions types from the
        connectivity information in the structure/topology input file and the
        bonded types specified in the configuration file.
    """
    if topol is not None:
        bonds_2, bonds_3, bonds_4 = prepare_index_based_bonds(
            molecules, topol
        )
    else:
        bonds_2, bonds_3, bonds_4, bb_index = prepare_bonds_old(
            molecules, names, bonds, indices, config
        )

    # Bonds
    bonds_2_atom1 = np.empty(len(bonds_2), dtype=int)
    bonds_2_atom2 = np.empty(len(bonds_2), dtype=int)
    bonds_2_equilibrium = np.empty(len(bonds_2), dtype=np.float64)
    bonds_2_strength = np.empty(len(bonds_2), dtype=np.float64)
    for i, b in enumerate(bonds_2):
        bonds_2_atom1[i] = b[0]
        bonds_2_atom2[i] = b[1]
        bonds_2_equilibrium[i] = b[2]
        bonds_2_strength[i] = b[3]

    # Angles
    bonds_3_atom1 = np.empty(len(bonds_3), dtype=int)
    bonds_3_atom2 = np.empty(len(bonds_3), dtype=int)
    bonds_3_atom3 = np.empty(len(bonds_3), dtype=int)
    bonds_3_equilibrium = np.empty(len(bonds_3), dtype=np.float64)
    bonds_3_strength = np.empty(len(bonds_3), dtype=np.float64)
    for i, b in enumerate(bonds_3):
        bonds_3_atom1[i] = b[0]
        bonds_3_atom2[i] = b[1]
        bonds_3_atom3[i] = b[2]
        bonds_3_equilibrium[i] = b[3]
        bonds_3_strength[i] = b[4]
    # Dihedrals
    bonds_4_atom1 = np.empty(len(bonds_4), dtype=int)
    bonds_4_atom2 = np.empty(len(bonds_4), dtype=int)
    bonds_4_atom3 = np.empty(len(bonds_4), dtype=int)
    bonds_4_atom4 = np.empty(len(bonds_4), dtype=int)
    number_of_coeff = 6
    len_of_coeff = 5
    bonds_4_coeff = np.empty(
        (len(bonds_4), number_of_coeff, len_of_coeff), dtype=np.float64
    )
    bonds_4_type = np.empty(len(bonds_4), dtype=int)
    bonds_4_last = np.zeros(len(bonds_4), dtype=int)
    for i, b in enumerate(bonds_4):
        bonds_4_atom1[i] = b[0]
        bonds_4_atom2[i] = b[1]
        bonds_4_atom3[i] = b[2]
        bonds_4_atom4[i] = b[3]
        bonds_4_coeff[i] = np.resize(b[4], (number_of_coeff, len_of_coeff))
        bonds_4_type[i] = b[5]
        if topol is not None:
            bonds_4_last[i] = b[6]
    if topol is None:
        bonds_4_last[bb_index] = 1

    return (
        bonds_2_atom1,
        bonds_2_atom2,
        bonds_2_equilibrium,
        bonds_2_strength,
        bonds_3_atom1,
        bonds_3_atom2,
        bonds_3_atom3,
        bonds_3_equilibrium,
        bonds_3_strength,  # noqa: E501
        bonds_4_atom1,
        bonds_4_atom2,
        bonds_4_atom3,
        bonds_4_atom4,
        bonds_4_coeff,
        bonds_4_type,
        bonds_4_last,  # noqa: E501
    )


def compute_bond_forces__plain(f_bonds, r, bonds_2, box_size):
    """Computes forces resulting from bonded interactions

    .. deprecated:: 1.0.0
        :code:`compute_bond_forces__plain` was replaced by compiled Fortran
        code prior to 1.0.0 release.
    """
    f_bonds.fill(0.0)
    energy = 0.0

    for i, j, r0, k in bonds_2:
        ri = r[i, :]
        rj = r[j, :]
        rij = rj - ri

        # Apply periodic boundary conditions to the distance rij
        for dim in range(len(rij)):
            rij[dim] -= box_size[dim] * np.around(rij[dim] / box_size[dim])
        dr = np.linalg.norm(rij)
        df = -k * (dr - r0)
        f_bond_vector = df * rij / dr
        f_bonds[i, :] -= f_bond_vector
        f_bonds[j, :] += f_bond_vector

        energy += 0.5 * k * (dr - r0) ** 2
    return energy


def compute_angle_forces__plain(f_angles, r, bonds_3, box_size):
    """Computes forces resulting from angular interactions

    .. deprecated:: 1.0.0
        :code:`compute_angle_forces__plain` was replaced by compiled Fortran
        code prior to 1.0.0 release.
    """
    f_angles.fill(0.0)
    energy = 0.0

    for a, b, c, theta0, k in bonds_3:
        ra = r[a, :] - r[b, :]
        rc = r[c, :] - r[b, :]

        for dim in range(len(ra)):
            ra[dim] -= box_size[dim] * np.around(ra[dim] / box_size[dim])
            rc[dim] -= box_size[dim] * np.around(rc[dim] / box_size[dim])

        xra = 1.0 / np.sqrt(np.dot(ra, ra))
        xrc = 1.0 / np.sqrt(np.dot(rc, rc))
        ea = ra * xra
        ec = rc * xrc

        cosphi = np.dot(ea, ec)
        theta = np.arccos(cosphi)
        xsinph = 1.0 / np.sqrt(1.0 - cosphi**2)

        d = theta - theta0
        f = -k * d

        xrasin = xra * xsinph * f
        xrcsin = xrc * xsinph * f

        fa = (ea * cosphi - ec) * xrasin
        fc = (ec * cosphi - ea) * xrcsin

        f_angles[a, :] += fa
        f_angles[c, :] += fc
        f_angles[b, :] -= fa + fc
        # f_angles[b, :] += -(fa + fc)

        energy -= 0.5 * f * d

    return energy


def compute_dihedral_forces__plain(f_dihedrals, r, bonds_4, box_size):
    """Computes forces resulting from dihedral interactions

    .. deprecated:: 1.0.0
        :code:`compute_dihedral_forces__plain` was replaced by compiled Fortran
        code prior to 1.0.0 release.
    """
    f_dihedrals.fill(0.0)
    energy = 0.0

    for a, b, c, d, coeffs, phase in bonds_4:
        f = r[a, :] - r[b, :]
        g = r[b, :] - r[c, :]
        h = r[d, :] - r[c, :]

        for dim in range(3):
            f -= box_size[dim] * np.around(f[dim] / box_size[dim])
            g -= box_size[dim] * np.around(g[dim] / box_size[dim])
            h -= box_size[dim] * np.around(h[dim] / box_size[dim])

        v = np.cross(f, g)
        w = np.cross(h, g)
        vv = np.dot(v, v)
        ww = np.dot(w, w)
        gn = np.linalg.norm(g)

        cosphi = np.dot(v, w)
        sinphi = np.dot(np.cross(v, w), g) / gn
        phi = np.arctan2(sinphi, cosphi)

        fg = np.dot(f, g)
        hg = np.dot(h, g)
        sc = v * fg / (vv * gn) - w * hg / (ww * gn)

        df = 0

        for m in range(len(coeffs[0])):
            energy += coeffs[0][m] * (1 + np.cos(m * phi - coeffs[1][m]))
            df += m * coeffs[0][m] * np.sin(m * phi - coeffs[1][m])

        force_on_a = df * gn * v / vv
        force_on_d = df * gn * w / ww

        f_dihedrals[a, :] -= force_on_a
        f_dihedrals[b, :] += df * sc + force_on_a
        f_dihedrals[c, :] -= df * sc + force_on_d
        f_dihedrals[d, :] += force_on_d
    return energy


def dipole_forces_redistribution(
    f_on_bead, f_dipoles, trans_matrices, a, b, c, d, type_array, last_bb
):
    """Redistribute electrostatic forces calculated from topologically
    reconstructed ghost dipole point charges to the backcone atoms of the protein.
    """
    f_on_bead.fill(0.0)
    for i, j, k, l, fd, matrix, dih_type, is_last in zip(
        a, b, c, d, f_dipoles, trans_matrices, type_array, last_bb
    ):
        if dih_type == 1:
            sum_force = fd[0] + fd[1]
            diff_force = fd[0] - fd[1]
            f_on_bead[i] += matrix[0] @ diff_force  # Atom A
            f_on_bead[j] += matrix[1] @ diff_force + 0.5 * sum_force  # Atom B
            f_on_bead[k] += matrix[2] @ diff_force + 0.5 * sum_force  # Atom C

            if is_last == 1:
                sum_force = fd[2] + fd[3]
                diff_force = fd[2] - fd[3]
                f_on_bead[j] += matrix[3] @ diff_force
                f_on_bead[k] += matrix[4] @ diff_force + 0.5 * sum_force
                f_on_bead[l] += matrix[5] @ diff_force + 0.5 * sum_force



"""Parses and handles the configuration information provided for the simulation
"""
import copy
import tomli
import datetime
import logging
import warnings
import numpy as np
from mpi4py import MPI
from dataclasses import dataclass, field
from typing import List, Union, ClassVar
from .force import Bond, Angle, Dihedral, Chi, Dielectric_type
from .barostat import Target_pressure
from .logger import Logger


@dataclass
class Config:
    """Configuration object

    Handles and verifies the simulation configuration specified in the
    configuration file.

    Attributes
    ----------
    gas_constant : float
        Constant value of the gas constant, R (equivalently the Boltzmann
        constant) in the units used internally in HyMD.
    coulomb_constant : float
        Constant value of the Coulomb constant which converts electric field
        values to forces and electric potential values to energies in the units
        used internally in HyMD.
    n_steps: int
        Number of time steps in the simulation.
    time_step: float
        *Outer* time step used in the simulation. If the rRESPA intergrator is
        used, the *inner* time step (the time step used in the integration of
        intramolecular bending, stretching, and torsional forces) is
        :code:`time_step / respa_inner`.
    box_size : list[float] or (D,) numpy.ndarray
        Simulation box size of simulation in :code:`D` dimensions in units of
        nanometers.
    mesh_size : list[int] or int or numpy.ndarray
        Number of grid points used for the discrete density grid.
    sigma : float
        Filter width representing the effective coarse-graining level of the
        particles in the simulation.
    kappa : float
        Compressibility parameter used in the relaxed incompressibility term in
        the interaction energy functional.
    n_print : int, optional
        Frequency of trajectory/energy output to the H5MD trajectory/energy
        output file (in units of number of time steps).
    tau : float, optional
        The time scale of the CSVR thermostat coupling.
    start_temperature : float, optional
        Generate starting temperature by assigning all particle velocities
        randomly according to the Maxwell-Boltzmann distribution at
        :code:`start_temperature` Kelvin prior to starting the simulation.
    target_temperature : float, optional
        Couple the system to a heat bath at :code:`target_temperature` Kelvin.
    mass : float, optional
        Mass of the particles in the simulation.
    hamiltonian : str, optional
        Specifies the interaction energy functional :math:`W[\\tilde\\phi]`
        for use with the particle-field interactions. Options:
        :code:`SquaredPhi`, :code:`DefaultNohChi`, or :code:`DefaultWithChi`.
    domain_decomposition : int, optional
        Specifies the interval (in time steps) of domain decomposition
        exchange, involving all MPI ranks sending and receiving particles
        according to the particles’ positions in the integration box and the
        MPI ranks’ assigned domain.
    integrator : str, optional
        Specifies the time integrator used in the simulation. Options:
        :code:`velocity-verlet` or :code:`respa`.
    respa_inner : int, optional
        The number of inner time steps in the rRESPA integrator. This denotes
        the number of intramolecular force calculations (stretching, bending,
        torsional) are performed between each impulse applied from the field
        forces.
    file_name : str, optional
        File path of the parsed configuration file.
    name : str, optional
        Name for the simulation.
    tags : list[str], optional
        Tags for the simulation.
    bonds : list[Bond], optional
        Specifies harmonic stretching potentials between particles in the
        same molecule.
    angle_bonds : list[Angle], optional
        Specifies harmonic angular bending potentials between particles in the
        same molecule.
    dihedrals : list[Dihedral], optional
        Specifies four-particle torsional potentials by cosine series.
    chi : list[Chi], optional
        Specifies :math:`\\chi`-interaction parameters between particle
        species.
    n_particles : int, optional
        Specifies the total number of particles in the input. Optional keyword
        for validation, ensuring the input HDF5 topology has the correct number
        of particles and molecules.
    max_molecule_size : int, optional
        Maximum size of any single molecule in the system. Used to speed up
        distribution of particles onto MPI ranks in a parallel fashion.
    n_flush : int, optional
        Frequency of HDF5 write buffer flush, forcing trajectory/energy to be
        written to disk (in units of number of :code:`n_print`).
    thermostat_work : float
        Work performed by the thermostat on the system.
    thermostat_coupling_groups : list[str], optional
        Specifies individual groups coupling independently to the CSVR
        thermostat. E.g. in a system containing :code:`"A"`, :code:`"B"`, and
        :code:`"C"` type particles,
        :code:`thermostat_coupling_groups = [["A", "B"], ["C"],]` would
        thermalise types :code:`"A"` and :code:`"B"` together and couple
        :code:`"C"` type particles to a different thermostat (all individual
        thermostats are at the same temperature, i.e. target_temperature
        Kelvin).
    initial_energy : float
        Value of the total energy prior to the start of the simulation.
    cancel_com_momentum : int, optional
        If :code:`True`, the total linear momentum of the center of mass is
        removed before starting the simulation. If an integer is specifed, the
        total linear momentum of the center of mass is removed every
        :code:`remove_com_momentum` time steps. If :code:`False`, the linear
        momentum is never removed.
    coulombtype : str, optional
        Specifies the type of electrostatic Coulomb interactions in the system.
        The strength of the electrostatic forces is modulated by the relative
        dielectric constant of the simulation medium, specified with the
        :code:`dielectric_const` keyword. Charges for individual particles are
        specified in the structure/topology HDF5 input file, not in the
        configuration file. If no charges (or peptide backbone dipoles) are
        present, the electrostatic forces will not be calculated even if this
        keyword is set to :code:`"PIC_Spectral"`.
    dielectric_const : float, optional
        Specifies the relative dielectric constant of the simulation medium
        which regulates the strength of the electrostatic interactions. When
        using helical propensity dihedrals, this keyword must be specified—even
        if electrostatics are not included with the :code:`coulombtype`
        keyword.
    dielectric_type: list[float], optional
        Specifies the relative dielectric constant of the simulation medium
        which regulates the strength of the electrostatic interactions. The list assigns
        relative dielectric values to each bead type, and an anisotropic
        dielectric function is obtained from a weighted average.


    See also
    --------
    hymd.input_parser.Bond :
        Two-particle bond type dataclass.
    hymd.input_parser.Angle :
        Three-particle bond type dataclass.
    hymd.input_parser.Dihedral :
        Four-particle bond type dataclass.
    """

    gas_constant: ClassVar[float] = 0.0083144621  # kJ mol-1 K-1
    coulomb_constant: ClassVar[float] = 138.935458  # kJ nm mol-1 e-2

    n_steps: int
    time_step: float
    mesh_size: Union[Union[List[int], np.ndarray], int]
    sigma: float
    kappa: float

    dtype: np.dtype = None
    box_size: Union[List[float], np.ndarray] = None
    n_print: int = None
    tau: float = None
    start_temperature: Union[float, bool] = None
    target_temperature: Union[float, bool] = None
    mass: float = None
    hamiltonian: str = None
    domain_decomposition: Union[int, bool] = None
    integrator: str = None
    respa_inner: int = 1
    file_name: str = "<config file path unknown>"
    name: str = None
    tags: List[str] = field(default_factory=list)
    bonds: List[Bond] = field(default_factory=list)
    angle_bonds: List[Angle] = field(default_factory=list)
    dihedrals: List[Dihedral] = field(default_factory=list)
    chi: List[Chi] = field(default_factory=list)
    n_particles: int = None
    max_molecule_size: int = None
    n_flush: int = None
    thermostat_work: float = 0.0
    thermostat_coupling_groups: List[List[str]] = field(default_factory=list)
    initial_energy: float = None
    cancel_com_momentum: Union[int, bool] = False
    coulombtype: str = None
    convergence_type: str = None
    pol_mixing: float = None
    dielectric_const: float = None
    conv_crit: float = None
    dielectric_type: List[Dielectric_type] = field(default_factory=list)
    self_energy: float = None
    type_charges: Union[List[float], np.ndarray] = None

    # For NPT runs
    rho0: float = None
    a: float = None
    pressure: bool = False
    barostat: str = None
    barostat_type: str = None
    tau_p: float = None
    target_pressure: List[Target_pressure] = field(default_factory=list)
    n_b: int = None
    m: List[float] = field(default_factory=list)

    def __str__(self):
        target_pressure_str = "\ttarget_pressure:\n" + "".join(
            "\t\tP_L: "
            + f"{self.target_pressure.P_L}\n"
            + "\t\tP_N: "
            + f"{self.target_pressure.P_N}\n"
        )
        bonds_str = "\tbonds:\n" + "".join(
            [
                (f"\t\t{k.atom_1} {k.atom_2}: " f"{k.equilibrium}, {k.strength}\n")
                for k in self.bonds
            ]
        )
        angle_str = "\tangle_bonds:\n" + "".join(
            [
                (
                    f"\t\t{k.atom_1} {k.atom_2} {k.atom_3}: "
                    + f"{k.equilibrium}, {k.strength}\n"
                )
                for k in self.angle_bonds
            ]
        )
        dihedrals_str = "\tdihedrals:\n" + "".join(
            [
                (
                    f"\t\t{k.atom_1} {k.atom_2} {k.atom_3} {k.atom_4}: "
                    # This might need to be fixed/made prettier, probably
                    # there's an easier way
                    + (
                        "\n\t\t"
                        + " " * len(f"{k.atom_1} {k.atom_2} {k.atom_3} {k.atom_4}: ")
                    ).join(
                        map(
                            str,
                            [
                                [round(num, 3) for num in c_in]
                                if isinstance(c_in, list)
                                else c_in
                                for c_in in k.coeffs
                            ],
                        )
                    )
                    + (
                        "\n\t\t"
                        + " " * len(f"{k.atom_1} {k.atom_2} {k.atom_3} {k.atom_4}: ")
                    )
                    + f"dih_type = {k.dih_type}\n"
                )
                for k in self.dihedrals
            ]
        )
        chi_str = "\tchi:\n" + "".join(
            [
                (f"\t\t{k.atom_1} {k.atom_2}: " + f"{k.interaction_energy}\n")
                for k in self.chi
            ]
        )

        """ If dielectric wanted as dictionary
            dielectric_type_str = "\tdielectric_type:\n" + "".join(
                    [
                        (f"\t\t{k.atom_1}: " + f"{k.dielectric_value}\n")
                        for k in self.dielectric_type
                    ]
                )

        """

        thermostat_coupling_groups_str = ""
        if any(self.thermostat_coupling_groups):
            thermostat_coupling_groups_str = (
                "\tthermostat_coupling_groups:\n"
                + "".join(
                    [
                        "\t\t" + ", ".join([f"{n}" for n in ng]) + "\n"
                        for ng in self.thermostat_coupling_groups
                    ]
                )
            )

        ret_str = f'\n\n\tConfig: {self.file_name}\n\t{50 * "-"}\n'
        for k, v in self.__dict__.items():
            if k not in (
                "target_pressure",
                "bonds",
                "angle_bonds",
                "dihedrals",
                "chi",
                "thermostat_coupling_groups",
            ):
                ret_str += f"\t{k}: {v}\n"
        ret_str += (
            target_pressure_str
            + bonds_str
            + angle_str
            + dihedrals_str
            + chi_str
            + thermostat_coupling_groups_str
        )
        return ret_str


def read_config_toml(file_path):
    with open(file_path, "r") as in_file:
        file_content = in_file.read()
    return file_content

def propensity_potential_coeffs(x: float, comm):
    alpha_coeffs = np.array(
        [
            [7.406, -5.298, -2.570, 1.336, 0.739],
            [-0.28632126, 1.2099146, 1.18122138, 0.49075168, 0.98495911],
        ]
    )
    beta_coeffs = np.array(
        [
            [3.770, 5.929, -4.151, -0.846, 0.190],
            [-0.2300693, -0.0583289, 0.99342396, 1.03237971, 2.90160988],
        ]
    )
    coil_coeffs = np.array(
        [
            [1.416, -0.739, 0.990, -0.397, 0.136],
            [1.3495933, 0.45649087, 2.30441057, -0.12274901, -0.26179939],
        ]
    )

    zero_add = np.zeros((2, 5))
    if x == -1:
        return np.concatenate((alpha_coeffs, zero_add))
    elif x == 0:
        return np.concatenate((coil_coeffs, zero_add))
    elif x == 1:
        return np.concatenate((beta_coeffs, zero_add))

    abs_x = np.abs(x)
    if abs_x > 1:
        err_str = (
            f"The provided value of λ = {x} is out of λ definition range, "
            f"[-1.0, 1.0]."
        )
        Logger.rank0.log(logging.ERROR, err_str)
        if comm.Get_rank() == 0:
            raise ValueError(err_str)

    else:
        coil_coeffs[0] *= 1 - abs_x
        if x < 0:
            alpha_coeffs[0] *= 0.5 * (abs_x - x)
            return np.concatenate((alpha_coeffs, coil_coeffs))
        else:
            beta_coeffs[0] *= 0.5 * (abs_x + x)
            return np.concatenate((beta_coeffs, coil_coeffs))


def parse_config_toml(toml_content, file_path=None, comm=MPI.COMM_WORLD):
    config_dict = {}

    # read toml to a dictionary
    toml_content = tomli.loads(toml_content)

    # Defaults = None
    for n in (
        "box_size",
        "n_print",
        "tau",
        "start_temperature",
        "target_temperature",
        "mass",
        "hamiltonian",
        "domain_decomposition",
        "integrator",
        "name",
        "n_particles",
        "max_molecule_size",
        "coulombtype",
        "convergence_type",
        "dielectric_const",
        "pol_mixing",
        "conv_crit",
        "dielectric_type",
        "n_b",
        "box_size",
        "n_flush",
        "self_energy",
        "dtype",
    ):
        config_dict[n] = None

    # Defaults = []

    for n in (
        "bonds",
        "angle_bonds",
        "dihedrals",
        "chi",
        "tags",
        "m",
        "dielectric_type",
        "target_pressure",
        "type_charges",
    ):
        config_dict[n] = []

    # Defaults: bool
    config_dict["pressure"] = False

    # Flatten the .toml dictionary, ignoring the top level [tag] directives (if
    # any).
    for k, v in toml_content.items():
        if isinstance(v, dict):
            if k == "nn": # Don't parse diff-hymd optimization options
                continue
            for nested_k, nested_v in v.items():
                config_dict[nested_k] = nested_v
        else:
            config_dict[k] = v
    for k, v in config_dict.items():
        if k == "bonds":
            config_dict["bonds"] = [None] * len(v)
            for i, b in enumerate(v):
                config_dict["bonds"][i] = Bond(
                    atom_1=b[0],
                    atom_2=b[1],
                    equilibrium=b[2],
                    strength=b[3],
                )
        if k == "angle_bonds":
            config_dict["angle_bonds"] = [None] * len(v)
            for i, b in enumerate(v):
                config_dict["angle_bonds"][i] = Angle(
                    atom_1=b[0],
                    atom_2=b[1],
                    atom_3=b[2],
                    equilibrium=b[3],
                    strength=b[4],
                )
        if k == "dihedrals":
            config_dict["dihedrals"] = [None] * len(v)
            for i, b in enumerate(v):
                try:
                    dih_type = int(b[2][0])
                except IndexError:
                    Logger.rank0.log(
                        logging.WARNING, "Dihedral type not provided, defaulting to 0."
                    )
                    dih_type = 0

                    # Probably it's better to move this in check_dihedrals?
                    wrong_len = len(b[1]) not in (1, 2)
                    wrong_type_1 = len(b[1]) == 1 and not isinstance(b[1][0], float)
                    wrong_type_2 = len(b[1]) == 2 and not isinstance(b[1][0], list)
                    if wrong_len or wrong_type_1 or wrong_type_2:
                        err_str = (
                            "The coefficients specified for the dihedral type "
                            "(0) do not match the correct structure. Either "
                            "use [lambda] or [[cn_prop], [dn_prop]], or select"
                            " the correct dihedral type."
                        )
                        Logger.rank0.log(logging.ERROR, err_str)
                        if comm.Get_rank() == 0:
                            raise RuntimeError(err_str)

                # FIXME: this is messy af, I don't like it
                if dih_type == 0 and isinstance(b[1][0], (float, int)):
                    coeff = propensity_potential_coeffs(b[1][0], comm)
                elif dih_type == 1 and len(b[1]) == 3:
                    coeff = np.array(
                        propensity_potential_coeffs(b[1][0][0], comm).tolist()
                        + b[1][1:]
                    )
                elif dih_type == 2:
                    coeff = np.array(b[1])
                else:
                    coeff = np.insert(np.array(b[1]), 2, np.zeros((2, 5)), axis=0)

                config_dict["dihedrals"][i] = Dihedral(
                    atom_1=b[0][0],
                    atom_2=b[0][1],
                    atom_3=b[0][2],
                    atom_4=b[0][3],
                    coeffs=coeff,
                    dih_type=dih_type,
                )
        if k == "chi":
            config_dict["chi"] = [None] * len(v)
            for i, c in enumerate(v):
                c_ = sorted([c[0], c[1]])
                config_dict["chi"][i] = Chi(
                    atom_1=c_[0], atom_2=c_[1], interaction_energy=c[2]
                )

        """
        if k == "dielectric_type":
            config_dict["dielectric_type"] = [None] * len(v)
            for i, c in enumerate(v):
                c_ = sorted([c[0][0]])
                config_dict["dielectric_type"][i] = Dielectric_type(
                    atom_1=c_[0], dielectric_value=c[1][0]
                )
        """
        if k == "target_pressure":
            if len(v) == 2:
                config_dict["target_pressure"] = Target_pressure(
                    P_L=v[0],
                    P_N=v[1],  # check condition # V array (still read as array?)
                )
            elif len(v) == 1:
                config_dict["target_pressure"] = Target_pressure(P_L=v[0], P_N=None)
            else:
                config_dict["target_pressure"] = Target_pressure(P_L=None, P_N=None)

    if file_path is not None:
        config_dict["file_name"] = file_path

    for n in (
        "n_steps", "time_step", "mesh_size", "sigma", "kappa"
    ):
        if n not in config_dict:
            err_str = (
                f"No {n} specified in config file {file_path}. Unable to start"
                f" simulation."
            )
            Logger.rank0.log(logging.ERROR, err_str)
            if comm.Get_rank() == 0:
                raise ValueError(err_str)
    return Config(**config_dict)


def check_n_particles(config, indices, comm=MPI.COMM_WORLD):
    n_particles = comm.allreduce(len(indices), MPI.SUM)
    if config.n_particles is None:
        config = copy.deepcopy(config)
        config.n_particles = n_particles
        info_str = (
            f"No n_particles found in toml file {config.file_name}, defaulting"
            f" to indices.shape ({n_particles})"
        )
        Logger.rank0.log(logging.INFO, info_str)
        return config

    if n_particles != config.n_particles:
        warn_str = (
            f"n_particles in {config.file_name} ({config.n_particles}) does "
            "not match the shape of the indices array in the .HDF5 file "
            f"({n_particles}). Defaulting to using indices.shape "
            f"({n_particles})"
        )
        Logger.rank0.log(logging.WARNING, warn_str)
        if comm.Get_rank() == 0:
            warnings.warn(warn_str)
        config = copy.deepcopy(config)
        config.n_particles = n_particles
    return config


def check_max_molecule_size(config, comm=MPI.COMM_WORLD):
    if config.max_molecule_size is None:
        info_str = (
            f"No max_molecule_size found in toml file {config.file_name}, "
            f"defaulting to 201"
        )
        Logger.rank0.log(logging.INFO, info_str)
        config = copy.deepcopy(config)
        config.max_molecule_size = 201
        return config

    if config.max_molecule_size < 1:
        warn_str = (
            f"max_molecule_size in {config.file_name} must be a positive "
            f"integer, not {config.max_molecule_size}, defaulting to 201"
        )
        Logger.rank0.log(logging.WARNING, warn_str)
        if comm.Get_rank() == 0:
            warnings.warn(warn_str)
        config = copy.deepcopy(config)
        config.max_molecule_size = 201
        return config
    return config


def _find_unique_names(config, names, comm=MPI.COMM_WORLD):
    unique_names = np.unique(names)
    receive_buffer = comm.gather(unique_names, root=0)

    gathered_unique_names = None
    if comm.Get_rank() == 0:
        gathered_unique_names = np.unique(np.concatenate(receive_buffer))
    unique_names = comm.bcast(gathered_unique_names, root=0)
    unique_names = sorted([n.decode("UTF-8") for n in unique_names])
    config.unique_names = unique_names
    config.n_types = len(unique_names)
    return config


def _setup_type_to_name_map(config, names, types, comm=MPI.COMM_WORLD):
    if not hasattr(config, "unique_names"):
        config = _find_unique_names(config, names)
    name_to_type_ = {}
    for n, t in zip(names, types):
        n = n.decode("utf-8")
        if n not in name_to_type_:
            name_to_type_[n] = t
    receive_buffer = comm.gather(name_to_type_, root=0)
    gathered_dict = None
    if comm.Get_rank() == 0:
        gathered_dict = {}
        for d in receive_buffer:
            for k, v in d.items():
                if k not in gathered_dict:
                    gathered_dict[k] = v
                else:
                    assert v == gathered_dict[k]
    name_to_type_map = comm.bcast(gathered_dict, root=0)
    config.name_to_type_map = name_to_type_map
    config.type_to_name_map = {v: k for k, v in name_to_type_map.items()}
    return config


def check_bonds(config, names, comm=MPI.COMM_WORLD):
    if not hasattr(config, "unique_names"):
        config = _find_unique_names(config, names)
    unique_names = config.unique_names

    for b in config.bonds:
        if b.atom_1 not in unique_names or b.atom_2 not in unique_names:
            missing_str = ""
            if b.atom_1 not in unique_names:
                if b.atom_2 not in unique_names:
                    if b.atom_1 == b.atom_2:
                        missing_str = f"no {b.atom_1} atoms"
                    else:
                        missing_str = f"neither {b.atom_1}, {b.atom_2} atoms"
                else:
                    missing_str = f"no {b.atom_1} atoms"
            else:
                missing_str = f"no {b.atom_2} atoms"

            warn_str = (
                f"Bond type {b.atom_1}--{b.atom_2} specified in "
                f"{config.file_name} but {missing_str} are present in the "
                f"specified system (names array)"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
    return config


def check_angles(config, names, comm=MPI.COMM_WORLD):
    if not hasattr(config, "unique_names"):
        config = _find_unique_names(config, names)
    unique_names = config.unique_names

    for a in config.angle_bonds:
        if (
            a.atom_1 not in unique_names
            or a.atom_2 not in unique_names
            or a.atom_3 not in unique_names
        ):
            missing = [
                a.atom_1 not in unique_names,
                a.atom_2 not in unique_names,
                a.atom_3 not in unique_names,
            ]
            missing_names = [
                atom
                for i, atom in enumerate([a.atom_1, a.atom_2, a.atom_3])
                if missing[i]
            ]
            missing_str = ", ".join(np.unique(missing_names))

            warn_str = (
                f"Angle bond type {a.atom_1}--{a.atom_2}--{a.atom_3} "
                f"specified in {config.file_name} but no {missing_str} atoms "
                f"are present in the specified system (names array)"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
    return config


def check_dihedrals(config, names, comm=MPI.COMM_WORLD):
    if not hasattr(config, "unique_names"):
        config = _find_unique_names(config, names)
    unique_names = config.unique_names

    for d in config.dihedrals:
        if (
            d.atom_1 not in unique_names
            or d.atom_2 not in unique_names
            or d.atom_3 not in unique_names
            or d.atom_4 not in unique_names
        ):
            missing = [
                d.atom_1 not in unique_names,
                d.atom_2 not in unique_names,
                d.atom_3 not in unique_names,
                d.atom_4 not in unique_names,
            ]
            missing_names = [
                atom
                for i, atom in enumerate([d.atom_1, d.atom_2, d.atom_3, d.atom_4])
                if missing[i]
            ]
            missing_str = ", ".join(np.unique(missing_names))

            warn_str = (
                f"Dihedral type {d.atom_1}--{d.atom_2}--{d.atom_3}--{d.atom_4}"
                f" specified in {config.file_name} but no {missing_str} atoms "
                f"are present in the specified system (names array)"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
    return config


def check_chi(config, names, comm=MPI.COMM_WORLD):
    if not hasattr(config, "unique_names"):
        config = _find_unique_names(config, names)
    unique_names = config.unique_names

    for c in config.chi:
        if c.atom_1 not in unique_names or c.atom_2 not in unique_names:
            missing_str = ""
            if c.atom_1 not in unique_names:
                if c.atom_2 not in unique_names:
                    if c.atom_1 == c.atom_2:
                        missing_str = f"no {c.atom_1} atoms"
                    else:
                        missing_str = f"neither {c.atom_1}, {c.atom_2} atoms"
                else:
                    missing_str = f"no {c.atom_1} atoms"
            else:
                missing_str = f"no {c.atom_2} atoms"

            warn_str = (
                f"Chi interaction {c.atom_1}--{c.atom_2} specified in "
                f"{config.file_name} but {missing_str} are present in the "
                f"specified system (names array)"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)

    for i, n in enumerate(unique_names):
        for m in unique_names[i + 1 :]:
            found = False
            for c in config.chi:
                if (c.atom_1 == n and c.atom_2 == m) or (
                    c.atom_1 == m and c.atom_2 == n
                ):
                    found = True
            if not found:
                config.chi.append(Chi(atom_1=n, atom_2=m, interaction_energy=0.0))
                warn_str = (
                    f"Atom types {n} and {m} found in the "
                    f"system, but no chi interaction {n}--{m} "
                    f"specified in {config.file_name}. Defaulting to "
                    f"chi[{n}, {m}] = 0"
                )
                Logger.rank0.log(logging.WARNING, warn_str)
                if comm.Get_rank() == 0:
                    warnings.warn(warn_str)
    return config


def check_box_size(config, input_box, comm=MPI.COMM_WORLD):
    if config.box_size is not None:
        config.box_size = np.array(config.box_size, dtype=np.float32)
        if input_box.all() and not np.allclose(config.box_size, input_box, atol=0.009):
            err_str = (
                f"Box size specified in {config.file_name}: "
                f"{config.box_size} does not match input box:"
                f"{input_box}"
            )
            Logger.rank0.log(logging.ERROR, err_str)
            if comm.Get_rank() == 0:
                raise ValueError(err_str)
    else:
        if input_box.all():
            config.box_size = input_box
        else:
            err_str = f"No box information found"
            Logger.rank0.log(logging.ERROR, err_str)
            if comm.Get_rank() == 0:
                raise ValueError(err_str)

    for b in config.box_size:
        if b <= 0.0:
            err_str = (
                f"Invalid box size specified in {config.file_name}: "
                f"{config.box_size}"
            )
            Logger.rank0.log(logging.ERROR, err_str)
            if comm.Get_rank() == 0:
                raise ValueError(err_str)
    return config


def check_integrator(config, comm=MPI.COMM_WORLD):
    if config.integrator.lower() not in ("velocity-verlet", "respa"):
        err_str = (
            f"Invalid integrator specified in {config.file_name}: "
            f'{config.integrator}. Available options "velocity-verlet" or '
            f'"respa".'
        )
        Logger.rank0.log(logging.ERROR, err_str)
        if comm.Get_rank() == 0:
            raise ValueError(err_str)

    if config.integrator.lower() == "respa":
        if isinstance(config.respa_inner, float):
            warn_str = (
                f"Number of inner rRESPA time steps in "
                f"{config.file_name}: {config.respa_inner} specified "
                f"as float, using {int(config.respa_inner)}"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
            config.respa_inner = int(config.respa_inner)
        elif not isinstance(config.respa_inner, int):
            err_str = (
                f"Invalid number of inner rRESPA time steps in "
                f"{config.file_name}: {config.respa_inner}. Must be "
                f"positive integer"
            )
            Logger.rank0.log(logging.ERROR, err_str)
            raise TypeError(err_str)

        if config.respa_inner <= 0:
            err_str = (
                f"Invalid number of inner rRESPA time steps in "
                f"{config.file_name}: {config.respa_inner}. Must be "
                f"positive integer"
            )
            Logger.rank0.log(logging.ERROR, err_str)
            raise ValueError(err_str)

    if config.integrator.lower() == "velocity-verlet" and config.respa_inner != 1:
        warn_str = (
            f"Integrator type Velocity-Verlet specified in {config.file_name} "
            f"and inner rRESPA time steps set to {config.respa_inner}. "
            f"Using respa_inner = 1"
        )
        Logger.rank0.log(logging.WARNING, warn_str)
        if comm.Get_rank() == 0:
            warnings.warn(warn_str)
        config.respa_inner = 1

    return config


def check_hamiltonian(config, comm=MPI.COMM_WORLD):
    valid_hamiltonians = ["defaultnochi", "defaultwithchi", "squaredphi"]

    if config.hamiltonian is None:
        if len(config.chi) == 0:
            warn_str = (
                f"No hamiltonian form and no chi interactions specified in "
                f"{config.file_name}, defaulting to DefaultNoChi hamiltonian"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
            config.hamiltonian = "DefaultNoChi"
        else:
            warn_str = (
                f"No hamiltonian form specified in {config.file_name}, but "
                f"chi interactions are specified, defaulting to "
                f"DefaultWithChi hamiltonian"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
            config.hamiltonian = "DefaultWithChi"
    elif config.hamiltonian.lower() not in valid_hamiltonians:
        err_str = (
            f"The specified Hamiltonian {config.hamiltonian} was not "
            f"recognized as a valid Hamiltonian."
        )
        Logger.rank0.log(logging.ERROR, err_str)
        raise NotImplementedError(err_str)

    return config


def check_n_flush(config, comm=MPI.COMM_WORLD):
    if config.n_print:
        if config.n_flush is None:
            config.n_flush = 10000 // config.n_print
    return config


def check_n_print(config, comm=MPI.COMM_WORLD):
    if (
        not isinstance(config.n_print, int)
        and not isinstance(config.n_print, float)
        and config.n_print is not None
    ):
        err_str = f"invalid value for n_print ({config.n_print})"
        Logger.rank0.log(logging.ERROR, err_str)
        raise RuntimeError(err_str)

    if config.n_print is None or config.n_print <= 0:
        config.n_print = False
    elif not isinstance(config.n_print, int):
        if isinstance(config.n_print, float):
            warn_str = (
                f"n_print is a float ({config.n_print}), not int, using "
                f"{int(round(config.n_print))}"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)

            config.n_print = int(round(config.n_print))
        else:
            err_str = f"invalid value for n_print ({config.n_print})"
            Logger.rank0.log(logging.ERROR, err_str)
            raise RuntimeError(err_str)
    return config


def check_tau(config, comm=MPI.COMM_WORLD):
    if config.tau is None and config.target_temperature is not None:
        warn_str = "target temp specified but no tau, defaulting 0.7"
        config.tau = 0.7
        Logger.rank0.log(logging.WARNING, warn_str)
        if comm.Get_rank() == 0:
            warnings.warn(warn_str)
    return config


def check_start_and_target_temperature(config, comm=MPI.COMM_WORLD):
    """Validate provided starting and target thermostat temperatures

    Assesses the provided temperature target and ensures it is a non-negative
    floating point number or :code:`False`. Ensures the starting temperature is
    a non-negative floating point number or :code:`False`.

    If the value for either is :code:`None`, the returned configuration object
    has the values defaulted to :code:`False` in each case.

    Parameters
    ----------
    config : Config
        Configuration object.

    Returns
    -------
    validated_config : Config
        Configuration object with validated :code:`target_temperature` and
        :code:`start_temperature`.
    """
    for t in ("start_temperature", "target_temperature"):
        if getattr(config, t) is not None:
            try:
                if getattr(config, t) < 0:
                    warn_str = (
                        f"{t} set to negative value ({getattr(config, t)}), "
                        f"defaulting to False"
                    )
                    setattr(config, t, False)
                    Logger.rank0.log(logging.WARNING, warn_str)
                    if comm.Get_rank() == 0:
                        warnings.warn(warn_str)
            except TypeError as e:
                err_str = (
                    f"Could not interpret {t} = {repr(getattr(config, t))} as "
                    f"a number."
                )
                raise TypeError(err_str) from e

    if config.start_temperature is None:
        config.start_temperature = False
    if config.target_temperature is None:
        config.target_temperature = False
    return config


def check_mass(config, comm=MPI.COMM_WORLD):
    if config.mass is None:
        info_str = "no mass specified, defaulting to 72.0"
        config.mass = 72.0
        Logger.rank0.log(logging.INFO, info_str)
    elif not isinstance(config.mass, int) and not isinstance(config.mass, float):
        err_str = (
            f"specified mass is invalid type {config.mass}, " f"({type(config.mass)})"
        )
        Logger.rank0.log(logging.ERROR, err_str)
        raise TypeError(err_str)
    elif config.mass < 0:
        err_str = f"invalid mass specified, {config.mass}"
        Logger.rank0.log(logging.ERROR, err_str)
        raise ValueError(err_str)
    else:
        config.mass = float(config.mass)

    return config


def check_domain_decomposition(config, comm=MPI.COMM_WORLD):
    dd = config.domain_decomposition
    if dd is None or dd is False:
        config.domain_decomposition = False
        return config

    if isinstance(dd, int):
        if dd < 0:
            warn_str = "negative domain_decomposition specified, using False"
            config.domain_decomposition = False
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
    elif isinstance(dd, float):
        if dd < 0.0:
            warn_str = "negative domain_decomposition specified, using False"
            config.domain_decomposition = False
        else:
            warn_str = (
                f"domain_decomposition ({config.domain_decomposition}) is not "
                f"an integer, using {int(round(dd))}"
            )
            config.domain_decomposition = int(round(dd))
        Logger.rank0.log(logging.WARNING, warn_str)
        if comm.Get_rank() == 0:
            warnings.warn(warn_str)
    else:
        err_str = (
            f"invalid value for domain_decomposition "
            f"({config.domain_decomposition}) use an integer"
        )
        Logger.rank0.log(logging.ERROR, err_str)
        raise ValueError(err_str)

    return config


def check_name(config, comm=MPI.COMM_WORLD):
    if config.name is None:
        root_current_time = ""
        if comm.Get_rank() == 0:
            root_current_time = datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        current_time = comm.bcast(root_current_time, root=0)

        if config.name is None:
            config.name = "sim" + current_time
    else:
        config.name = str(config.name)
    return config


def check_NPT_conditions(config, comm=MPI.COMM_WORLD):
    """
    Check validity of barostat_type, barostat,
    a, rho0, target_pressure, tau_p
    """
    if config.barostat is None:
        if (
            config.tau_p is not None
            or (config.target_pressure.P_L and config.target_pressure.P_N) is not None
        ):
            err_str = (
                "barostat not specified but config.tau_p "
                "or config.target_pressure specified, cannot start simulation {config.barostat}"
            )
            Logger.rank0.log(logging.ERROR, err_str)
            if comm.Get_rank() == 0:
                raise TypeError(err_str)
        if config.a:
            warn_str = "a specified but no barostat," "setting a to average density"
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
        if config.rho0:
            warn_str = (
                "rho0 specified but no barostat," "setting rho0 to average density"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)

    if config.barostat is not None:
        if not config.barostat_type:
            config.barostat_type = "berendsen"
            warn_str = "barostat_type not specified," "setting to berendsen"
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
        if config.barostat != "isotropic" and config.barostat != "semiisotropic":
            err_str = "barostat option not recognised. Valid options: isotropic, semiisotropic"
            Logger.rank0.log(logging.ERROR, err_str)
            if comm.Get_rank() == 0:
                raise TypeError(err_str)
        if config.target_pressure.P_L is None:
            config.target_pressure.P_L = 1.0  # bar
            config.target_pressure.P_N = None
            if config.barostat == "semiisotropic":
                config.target_pressure.P_N = 1.0  # bar
            warn_str = (
                "barostat specified but no target_pressure, defaulting to 1.0 bar"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
        if config.tau_p is None:
            if config.tau <= 0.1:
                config.tau_p = config.tau * 10.0
            else:
                config.tau_p = 1.0
            warn_str = "barostat specified but no tau_p, defaulting to " + str(
                config.tau_p
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
        if not config.a:
            err_str = "a not specified; cannot start simulation {config.a}"
            Logger.rank0.log(logging.ERROR, err_str)
            if comm.Get_rank() == 0:
                raise TypeError(err_str)
        if not config.rho0:
            warn_str = "rho0 not specified;" "setting rho0 to average density"
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)

    return config


def check_thermostat_coupling_groups(config, comm=MPI.COMM_WORLD):
    if any(config.thermostat_coupling_groups):
        found = [0 for _ in config.unique_names]
        unique_names = [n for n in config.unique_names]
        for i, group in enumerate(config.thermostat_coupling_groups):
            for species in group:
                try:
                    ind = unique_names.index(species)
                    found[ind] += 1
                except ValueError as e:
                    err_str = (
                        f"Particle species {species} specified in thermostat "
                        f"coupling group {i}, but no {species} particles were "
                        "found in the system."
                    )
                    raise ValueError(err_str) from e
        if any([True if f > 1 else False for f in found]):
            for i, f in enumerate(found):
                if f > 1:
                    species = unique_names[i]
                    err_str = (
                        f"Particle species {species} specified in multiple "
                        "thermostat coupling groups; "
                        f"{config.thermostat_coupling_groups}."
                    )
                    raise ValueError(err_str)
        if not all([True if f == 1 else False for f in found]):
            for i, f in enumerate(found):
                if f != 1:
                    species = unique_names[i]
                    err_str = (
                        f"Particle species {species} not specified in any "
                        f"thermostat coupling group, but {species} particles "
                        "were found in the system"
                    )
                    raise ValueError(err_str)
    return config


def check_m(config, comm=MPI.COMM_WORLD):
    if config.m == []:
        config.m = [1.0 for t in range(config.n_types)]
    return config


def check_n_b(config, comm=MPI.COMM_WORLD):
    if config.n_b is None:
        warn_str = f"config.n_b not specified." "Defaulting to 1"
        config.n_b = 1
        Logger.rank0.log(logging.WARNING, warn_str)
        if comm.Get_rank() == 0:
            warnings.warn(warn_str)
    return config


def check_cancel_com_momentum(config, comm=MPI.COMM_WORLD):
    if isinstance(config.cancel_com_momentum, int):
        if config.cancel_com_momentum == 0 or config.cancel_com_momentum < 0:
            config.cancel_com_momentum = False
    elif not isinstance(config.cancel_com_momentum, bool):
        err_str = (
            f"Could not interpret {config.cancel_com_momentum} as an integer "
            f"or boolean."
        )
        raise ValueError(err_str)
    return config


def sort_dielectric_by_type_id(config, charges, types):
    """
    Creates a list of length N CG-particles, sorted after the charges from
    the input HDF5 file. Used in file_io.py
    """
    dielectric_val = np.zeros(config.n_types)
    for j in range(config.n_types):
        name = config.dielectric_type[j][0][0]
        type_id = config.name_to_type_map[name]
        dielectric_val[type_id] = config.dielectric_type[j][1][0]
    config.dielectric_type = dielectric_val.copy()

    N = len(charges)
    len_list = np.zeros(config.n_types)
    dielectric_by_types = np.zeros(N)
    for i in range(N):
        dielectric_by_types[i] = dielectric_val[types[i]]
    # print("types ", types)
    # print("diel  val", dielectric_by_types)
    # print("charges", charges)
    # print(config.name_to_type_map)
    return dielectric_by_types  # by types with each particle id


def check_charges_types_list(config, types, charges, comm=MPI.COMM_WORLD):
    """
    Creates a list of charge values of length types.
    Charges are sorted according to type ID. Used in field.py.
    # TODO: this is messy, we should fix it
    """
    if charges is None:
        config.type_charges = [0.0] * config.n_types
        return config

    check_val = -100.0  # some random value that will never be encountered
    charges_list = np.full(config.n_types, check_val)  # gatherv cant handle None
    rank = comm.Get_rank()
    # print(charges_list)
    for t_ in range(config.n_types):
        if t_ in types:
            charges_list[t_] = charges[types == t_][0]

    nprocs = int(comm.Get_size())
    recv_charges = None
    if rank == 0:
        recv_charges = np.full(
            config.n_types * nprocs, check_val
        )  # gatherv cant handle None
    comm.Gather(charges_list, recv_charges, root=0)

    ## make a charges list
    if rank == 0:
        config_charges = np.zeros(config.n_types)
        test_config = np.full(config.n_types, check_val)
        for j in range(nprocs):
            for i in range(config.n_types):
                if recv_charges[i + j * config.n_types] != check_val:
                    config_charges[i] = recv_charges[i + j * config.n_types]
            if np.any([test_config, config_charges]):
                continue
            else:
                break

        # print(config_charges)
        # print(config.name_to_type_map)
        # print(charges[types == 0][0],charges[types == 1][0],charges[types == 2][0],charges[types == 3][0],
        #        charges[types == 4][0])
    else:
        config_charges = np.zeros(config.n_types)

    comm.Bcast(config_charges, root=0)
    # print("config_charges", config_charges , "rank", rank)
    # print(recv_charges)
    # rank = comm.Get_rank()
    # print(all_charges)
    config.type_charges = config_charges
    return config


def check_dielectric(config, comm=MPI.COMM_WORLD):
    """
    Error handling for electrostatics.
    Unit testing of toml/tomli input.
    """
    err_str_const = "Dielectric constant not given."
    if config.coulombtype == "PIC_Spectral":
        assert config.dielectric_const != None, err_str_const

    err_str = "Dielectric list is empty."
    if config.coulombtype == "PIC_Spectral_GPE":
        assert len(config.dielectric_type) != 0, err_str
        # default values
        if config.pol_mixing is None:
            config.pol_mixing = 0.6
        if config.conv_crit is None:
            config.conv_crit = 1e-6
    return config


def check_charges(config, charges, comm=MPI.COMM_WORLD):
    """Check if charges across ranks sum to zero.

    Parameters
    ----------
    charges : (N,) numpy.ndarray
        Array of floats with charges for :code:`N` particles.
    comm : mpi4py.Comm, optional
        MPI communicator, defaults to :code:`mpi4py.COMM_WORLD`.
    """
    total_charge = comm.allreduce(np.sum(charges), MPI.SUM)

    if not np.isclose(total_charge, 0.0):
        warn_str = (
            f"Charges in the input file do not sum to zero. "
            f"Total charge is {total_charge}."
        )
        Logger.rank0.log(logging.WARNING, warn_str)
        if comm.Get_rank() == 0:
            warnings.warn(warn_str)

    # if charges are ok, compute self energy
    if config.coulombtype == "PIC_Spectral":
        from .field import compute_self_energy_q

        config.self_energy = compute_self_energy_q(config, charges, comm=comm)

    return config


def check_config(
    config, indices, names, types, charges, input_box, comm=MPI.COMM_WORLD
):
    """Performs various checks on the specfied config to ensure consistency

    Parameters
    ----------
    config : Config
        Configuration object.
    indices : (N,) numpy.ndarray
        Array of integer indices for :code:`N` particles.
    names : (N,) numpy.ndarray
        Array of string names for :code:`N` particles.
    types : (N,) numpy.ndarray
        Array of integer type indices for :code:`N` particles.
    charges : (N,) numpy.ndarray
        Array of floats charges for :code:`N` particles.
    comm : mpi4py.Comm, optional
        MPI communicator, defaults to :code:`mpi4py.COMM_WORLD`.

    Returns
    -------
    config : Config
        Validated configuration object.
    """
    config = _find_unique_names(config, names, comm=comm)
    if types is not None:
        config = _setup_type_to_name_map(config, names, types, comm=comm)
    config = check_box_size(config, input_box, comm=comm)
    config = check_integrator(config, comm=comm)
    config = check_max_molecule_size(config, comm=comm)
    config = check_tau(config, comm=comm)
    config = check_start_and_target_temperature(config, comm=comm)
    config = check_mass(config, comm=comm)
    config = check_domain_decomposition(config, comm=comm)
    config = check_name(config, comm=comm)
    config = check_n_particles(config, indices, comm=comm)
    config = check_chi(config, names, comm=comm)
    config = check_bonds(config, names, comm=comm)
    config = check_angles(config, names, comm=comm)
    config = check_dihedrals(config, names, comm=comm)
    config = check_hamiltonian(config, comm=comm)
    config = check_NPT_conditions(config, comm=comm)
    config = check_n_b(config, comm=comm)
    config = check_m(config, comm=comm)
    config = check_thermostat_coupling_groups(config, comm=comm)
    config = check_cancel_com_momentum(config, comm=comm)
    config = check_dielectric(config, comm=comm)
    config = check_n_print(config, comm=comm)
    config = check_n_flush(config, comm=comm)
    config = check_charges_types_list(config, types, charges, comm=comm)
    if charges is not None:
        config = check_charges(config, charges, comm=comm)

    return config



"""Main simulation driver module
"""
import datetime
import h5py
import logging
import sys
from mpi4py import MPI
import numpy as np
import pmesh.pm as pmesh
import warnings
from .configure_runtime import configure_runtime
from .hamiltonian import get_hamiltonian
from .input_parser import check_config, sort_dielectric_by_type_id
from .logger import Logger, format_timedelta
from .file_io import distribute_input, OutDataset, store_static, store_data
from .field import (
    compute_field_force,
    update_field,
    compute_field_and_kinetic_energy,
    domain_decomposition,
    update_field_force_q,
    update_field_force_q_GPE,
    compute_field_energy_q_GPE,
    initialize_pm,
)
from .thermostat import (
    csvr_thermostat,
    cancel_com_momentum,
    generate_initial_velocities,
)
from .force import dipole_forces_redistribution, prepare_bonds
from .integrator import integrate_velocity, integrate_position
from .pressure import comp_pressure
from .plumed import PlumedBias


def main():
    """Main simulation driver

    Initializes structure, topology, and simulation configuration and iterates
    the molecular dynamics loop.
    """
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    if rank == 0:
        start_time = datetime.datetime.now()

    args, config, prng, topol = configure_runtime(sys.argv[1:], comm)

    if args.double_precision:
        dtype = np.float64
        config.dtype = dtype
        from .force import compute_bond_forces__fortran__double as compute_bond_forces
        from .force import compute_angle_forces__fortran__double as compute_angle_forces
        from .force import (
            compute_dihedral_forces__fortran__double as compute_dihedral_forces,
        )
    else:
        dtype = np.float32
        config.dtype = dtype
        from .force import compute_bond_forces__fortran as compute_bond_forces
        from .force import compute_angle_forces__fortran as compute_angle_forces
        from .force import compute_dihedral_forces__fortran as compute_dihedral_forces

    # read input .hdf5
    driver = "mpio" if not args.disable_mpio else None
    _kwargs = {"driver": driver, "comm": comm} if not args.disable_mpio else {}
    with h5py.File(args.input, "r", **_kwargs) as in_file:
        rank_range, molecules_flag = distribute_input(
            in_file,
            rank,
            size,
            config.n_particles,
            config.max_molecule_size if config.max_molecule_size else 201,
            comm=comm,
        )
        indices = in_file["indices"][rank_range]
        positions = in_file["coordinates"][-1, rank_range, :]
        positions = positions.astype(dtype)
        if "velocities" in in_file:
            velocities = in_file["velocities"][-1, rank_range, :]
            velocities = velocities.astype(dtype)
        else:
            velocities = np.zeros_like(positions, dtype=dtype)

        names = in_file["names"][rank_range]

        if "box" in in_file.attrs:
            config.box_size = np.array(in_file.attrs["box"])
        else:
            if getattr(config, "box_size") is None:
                err_str = (
                    f"No box size present in either config or input file. Unable to start"
                    f" simulation."
                )
                Logger.rank0.log(logging.ERROR, err_str)
                if comm.Get_rank() == 0:
                    raise ValueError(err_str)

        types = None
        bonds = None
        charges = None
        molecules = []
        if "types" in in_file:
            types = in_file["types"][rank_range]
        if molecules_flag:
            molecules = in_file["molecules"][rank_range]
            if topol is None:
                bonds = in_file["bonds"][rank_range]
        if "charge" in in_file:
            charges = in_file["charge"][rank_range]
            charges_flag = True
        else:
            charges_flag = False

        if "box" in in_file:
            input_box = in_file["box"][:]
        else:
            input_box = np.array([None, None, None])

    # finishes config setup and checks
    config = check_config(config, indices, names, types, charges, input_box, comm=comm)

    # import barostat if necessary
    if config.barostat_type == "berendsen":
        from .barostat import isotropic, semiisotropic
    elif config.barostat_type == "scr":
        from .barostat_scr import isotropic, semiisotropic

    # dielectric from toml
    if config.coulombtype == "PIC_Spectral_GPE":
        dielectric_sorted = sort_dielectric_by_type_id(config, charges, types)
        dielectric_flag = True

        if config.convergence_type is not None:
            if config.convergence_type == "csum":

                def conv_fun(comm, diff_mesh):
                    return diff_mesh.csum()  # check if allreduce needed

            elif config.convergence_type == "euclidean_norm":

                def conv_fun(comm, diff_mesh):
                    return diff_mesh.cnorm()  # check if allreduce nedded

            elif config.convergence_type == "max_diff":

                def conv_fun(comm, diffmesh):
                    msg = np.max(diffmesh)
                    res = comm.allreduce(sendobj=msg, op=MPI.MAX)
                    return res

        else:  # default choice is the max difference if not specified
            config.convergence_type = "max_diff"

            def conv_fun(comm, diffmesh):
                msg = np.max(diffmesh)
                res = comm.allreduce(sendobj=msg, op=MPI.MAX)
                return res

    else:
        dielectric_flag = False

    if config.start_temperature:
        velocities = generate_initial_velocities(velocities, config, prng, comm=comm)
    elif config.cancel_com_momentum:
        velocities = cancel_com_momentum(velocities, config, comm=comm)

    # set all PRNG to the root's PRNG
    # this is done to avoid communication between ranks in the thermostat
    prng = comm.bcast(prng, root=0)

    bond_forces = np.zeros_like(positions)
    angle_forces = np.zeros_like(positions)
    dihedral_forces = np.zeros_like(positions)
    reconstructed_forces = np.zeros_like(positions)
    field_forces = np.zeros_like(positions)
    elec_forces = np.zeros_like(positions)
    plumed_forces = np.zeros_like(positions)

    positions = np.mod(positions, config.box_size[None, :])

    # Initialize dipoles, populate them if protein_flag == True
    if args.disable_dipole:
        dipole_flag = 0
    else:
        dipole_flag = 1
    protein_flag = 0
    dipole_positions = np.zeros(shape=(4, 3), dtype=dtype)
    dipole_forces = np.zeros(shape=(4, 3))
    dipole_charges = np.zeros(shape=4)
    transfer_matrices = np.zeros(shape=(6, 3, 3), dtype=dtype)

    # Initialize energies
    field_energy = 0.0
    bond_energy = 0.0
    angle_energy = 0.0
    dihedral_energy = 0.0
    kinetic_energy = 0.0
    field_q_energy = 0.0
    plumed_bias = 0.0

    # more initialization
    bond_pr_ = np.zeros(3, dtype=dtype)
    angle_pr_ = np.zeros(3, dtype=dtype)

    hamiltonian = get_hamiltonian(config)

    pm_objs = initialize_pm(pmesh, config, comm)
    pm, field_list, elec_common_list, coulomb_list = pm_objs
    (
        phi,
        phi_fourier,
        force_on_grid,
        v_ext_fourier,
        v_ext,
        phi_transfer,
        phi_laplacian,
    ) = field_list

    # when electrostatics is not used, these objs are None
    phi_q, phi_q_fourier, psi, elec_field = elec_common_list

    if len(coulomb_list) == 2:
        (
            elec_field_fourier,
            psi_fourier,
        ) = coulomb_list

    elif len(coulomb_list) == 12:
        [
            phi_eps,
            phi_eps_fourier,
            phi_eta,
            phi_eta_fourier,
            phi_pol,
            phi_pol_prev,
            elec_dot,
            elec_field_contrib,
            Vbar_elec,
            Vbar_elec_fourier,
            force_mesh_elec,
            force_mesh_elec_fourier,
        ] = coulomb_list

    Logger.rank0.log(logging.INFO, f"pfft-python processor mesh: {str(pm.np)}")

    args_in = [
        velocities,
        indices,
        bond_forces,
        angle_forces,
        dihedral_forces,
        reconstructed_forces,
        field_forces,
        names,
        types,
    ]

    if charges_flag:
        args_in.append(charges)
        args_in.append(elec_forces)
    if dielectric_flag:
        args_in.append(dielectric_sorted)
    if args.plumed:
        args_in.append(plumed_forces)

    if config.domain_decomposition:
        (
            positions,
            velocities,
            indices,
            bond_forces,
            angle_forces,
            dihedral_forces,
            reconstructed_forces,
            field_forces,
            names,
            types,
            *optional,
        ) = domain_decomposition(
            positions,
            pm,
            *tuple(args_in),
            molecules=molecules if molecules_flag else None,
            bonds=bonds if molecules_flag else None,
            verbose=args.verbose,
            comm=comm,
        )

        if charges_flag:
            charges = optional.pop(0)
            elec_forces = optional.pop(0)
        if dielectric_flag:
            dielectric_sorted = optional.pop(0)
        if args.plumed:
            plumed_forces = optional.pop(0)
        if molecules_flag:
            bonds = optional.pop(0)
            molecules = optional.pop(0)

        # rebuild args_in to point to correct arrays
        args_in = [
            velocities,
            indices,
            bond_forces,
            angle_forces,
            dihedral_forces,
            reconstructed_forces,
            field_forces,
            names,
            types,
        ]

        if charges_flag:
            args_in.append(charges)
            args_in.append(elec_forces)
        if dielectric_flag:
            args_in.append(dielectric_sorted)
        if args.plumed:
            args_in.append(plumed_forces)

    if not args.disable_field:
        layouts = [
            pm.decompose(positions[types == t]) for t in range(config.n_types)
        ]  # noqa: E501
        update_field(
            phi,
            phi_laplacian,
            phi_transfer,
            layouts,
            force_on_grid,
            hamiltonian,
            pm,
            positions,
            types,
            config,
            v_ext,
            phi_fourier,
            v_ext_fourier,
            config.m,
            compute_potential=True,
        )
        (
            field_energy,
            kinetic_energy,
            field_q_energy,
        ) = compute_field_and_kinetic_energy(
            phi,
            phi_q,
            psi,
            velocities,
            hamiltonian,
            positions,
            types,
            v_ext,
            config,
            layouts,
            comm=comm,
        )
        compute_field_force(
            layouts, positions, force_on_grid, field_forces, types, config.n_types
        )
    else:
        kinetic_energy = comm.allreduce(0.5 * config.mass * np.sum(velocities**2))

    if charges_flag:
        layout_q = pm.decompose(positions)
        if config.coulombtype == "PIC_Spectral_GPE":  # dielectric_flag
            Vbar_elec, phi_eps, elec_dot = update_field_force_q_GPE(
                conv_fun,
                phi,
                types,
                charges,
                phi_q,
                phi_q_fourier,
                phi_eps,
                phi_eps_fourier,
                phi_eta,
                phi_eta_fourier,
                phi_pol_prev,
                phi_pol,
                elec_field,
                elec_forces,
                elec_field_contrib,
                psi,
                Vbar_elec,
                Vbar_elec_fourier,
                force_mesh_elec,
                force_mesh_elec_fourier,
                hamiltonian,
                layout_q,
                layouts,
                pm,
                positions,
                config,
                comm=comm,
            )

            field_q_energy = compute_field_energy_q_GPE(
                config, phi_eps, field_q_energy, elec_dot, comm=comm
            )

        if config.coulombtype == "PIC_Spectral":
            update_field_force_q(
                charges,
                phi_q,
                phi_q_fourier,
                psi,
                psi_fourier,
                elec_field_fourier,
                elec_field,
                elec_forces,
                layout_q,
                hamiltonian,
                pm,
                positions,
                config,
            )

    if molecules_flag:
        if not (
            args.disable_bonds and args.disable_angle_bonds and args.disable_dihedrals
        ):
            bonds_prep = prepare_bonds(molecules, names, bonds, indices, config, topol)
            (
                # two-particle bonds
                bonds_2_atom1,
                bonds_2_atom2,
                bonds_2_equilibrium,
                bonds_2_stength,
                # three-particle bonds
                bonds_3_atom1,
                bonds_3_atom2,
                bonds_3_atom3,
                bonds_3_equilibrium,
                bonds_3_stength,
                # four-particle bonds
                bonds_4_atom1,
                bonds_4_atom2,
                bonds_4_atom3,
                bonds_4_atom4,
                bonds_4_coeff,
                bonds_4_type,
                bonds_4_last,
            ) = bonds_prep

            bonds_4_coeff = np.asfortranarray(bonds_4_coeff)

            if bonds_4_type.any() > 1:
                err_str = (
                    "0 and 1 are the only currently supported dihedral angle " "types."
                )
                Logger.rank0.log(logging.ERROR, err_str)
                if rank == 0:
                    raise NotImplementedError(err_str)

            # Check if we have a protein
            protein_flag = comm.allreduce(bonds_4_type.any() == 1)
            if protein_flag and not args.disable_dipole:
                # Each rank will have different n_tors, we do not need to
                # domain decompose dipoles
                n_tors = len(bonds_4_atom1)
                dipole_positions = np.zeros((n_tors, 4, 3), dtype=dtype)
                # 4 because we need to take into account the last angle in the
                # molecule
                dipole_charges = np.array(
                    [
                        2 * [0.25, -0.25]
                        if (bonds_4_type[i], bonds_4_last[i]) == (1, 1)
                        else [0.25, -0.25, 0.0, 0.0]
                        if bonds_4_type[i] == 1
                        else 2 * [0.0, 0.0]
                        for i in range(n_tors)
                    ],
                    dtype=dtype,
                ).flatten()
                dipole_forces = np.zeros_like(dipole_positions)
                transfer_matrices = np.zeros(shape=(n_tors, 6, 3, 3), dtype=dtype)
                phi_dipoles = pm.create("real", value=0.0)
                phi_dipoles_fourier = pm.create("complex", value=0.0)
                psi_dipoles = pm.create("real", value=0.0)
                psi_dipoles_fourier = pm.create("complex", value=0.0)
                _SPACE_DIM = 3
                dipoles_field_fourier = [
                    pm.create("complex", value=0.0) for _ in range(_SPACE_DIM)
                ]
                dipoles_field = [
                    pm.create("real", value=0.0) for _ in range(_SPACE_DIM)
                ]

        positions = np.asfortranarray(positions)
        velocities = np.asfortranarray(velocities)
        bond_forces = np.asfortranarray(bond_forces)
        angle_forces = np.asfortranarray(angle_forces)
        dihedral_forces = np.asfortranarray(dihedral_forces)
        dipole_positions = np.asfortranarray(dipole_positions)
        transfer_matrices = np.asfortranarray(transfer_matrices)

        if not args.disable_bonds:
            bond_energy_, bond_pr_ = compute_bond_forces(
                bond_forces,
                positions,
                config.box_size,
                bonds_2_atom1,
                bonds_2_atom2,
                bonds_2_equilibrium,
                bonds_2_stength,
            )
            bond_energy = comm.allreduce(bond_energy_, MPI.SUM)
        else:
            bonds_2_atom1, bonds_2_atom2 = [], []
        if not args.disable_angle_bonds:
            angle_energy_, angle_pr_ = compute_angle_forces(
                angle_forces,
                positions,
                config.box_size,
                bonds_3_atom1,
                bonds_3_atom2,
                bonds_3_atom3,
                bonds_3_equilibrium,
                bonds_3_stength,
            )
            angle_energy = comm.allreduce(angle_energy_, MPI.SUM)
            # angle_pr = comm.allreduce(angle_pr_, MPI.SUM)

        if not args.disable_dihedrals:
            dihedral_energy_ = compute_dihedral_forces(
                dihedral_forces,
                positions,
                dipole_positions,
                transfer_matrices,
                config.box_size,
                bonds_4_atom1,
                bonds_4_atom2,
                bonds_4_atom3,
                bonds_4_atom4,
                bonds_4_coeff,
                bonds_4_type,
                bonds_4_last,
                dipole_flag,
            )
            dihedral_energy = comm.allreduce(dihedral_energy_, MPI.SUM)

        if protein_flag and not args.disable_dipole:
            dipole_positions = np.reshape(dipole_positions, (4 * n_tors, 3))
            dipole_forces = np.reshape(dipole_forces, (4 * n_tors, 3))

            layout_dipoles = pm.decompose(dipole_positions)
            update_field_force_q(
                dipole_charges,
                phi_dipoles,
                phi_dipoles_fourier,
                psi_dipoles,
                psi_dipoles_fourier,
                dipoles_field_fourier,
                dipoles_field,
                dipole_forces,
                layout_dipoles,
                hamiltonian,
                pm,
                dipole_positions,
                config,
            )

            dipole_positions = np.reshape(dipole_positions, (n_tors, 4, 3))
            dipole_forces = np.reshape(dipole_forces, (n_tors, 4, 3))

            dipole_positions = np.asfortranarray(dipole_positions)
            dipole_forces_redistribution(
                reconstructed_forces,
                dipole_forces,
                transfer_matrices,
                bonds_4_atom1,
                bonds_4_atom2,
                bonds_4_atom3,
                bonds_4_atom4,
                bonds_4_type,
                bonds_4_last,
            )

    else:
        bonds_2_atom1, bonds_2_atom2 = [], []

    config.initial_energy = (
        field_energy
        + kinetic_energy
        + bond_energy
        + angle_energy
        + dihedral_energy
        + field_q_energy
    )

    out_dataset = OutDataset(
        args.destdir,
        config,
        double_out=args.double_output,
        disable_mpio=args.disable_mpio,
    )

    store_static(
        out_dataset,
        rank_range,
        names,
        types,
        indices,
        config,
        bonds_2_atom1,
        bonds_2_atom2,
        molecules=molecules if molecules_flag else None,
        velocity_out=args.velocity_output,
        force_out=args.force_output,
        charges=charges if charges_flag else False,
        dielectrics=dielectric_sorted if dielectric_flag else False,
        plumed_out=True if args.plumed else False,
        comm=comm,
    )

    if config.n_print > 0:
        step = 0
        frame = 0
        if not args.disable_field:
            (
                field_energy,
                kinetic_energy,
                field_q_energy,
            ) = compute_field_and_kinetic_energy(
                phi,
                phi_q,
                psi,
                velocities,
                hamiltonian,
                positions,
                types,
                v_ext,
                config,
                layouts,
                comm=comm,
            )
        else:
            kinetic_energy = comm.allreduce(0.5 * config.mass * np.sum(velocities**2))

        temperature = (
            (2 / 3)
            * kinetic_energy
            / (config.gas_constant * config.n_particles)  # noqa: E501
        )

        if config.pressure or config.barostat:
            pressure = comp_pressure(
                phi,
                phi_q,
                psi,
                hamiltonian,
                velocities,
                config,
                phi_fourier,
                phi_laplacian,
                phi_transfer,
                positions,
                bond_pr_,
                angle_pr_,
                comm=comm,
            )

            # if rank ==0 : print(pressure[9:12])
            # print('phi_fft after pressure call: phi_fft[d=0]',phi_fft[0].value[0][0][0:2])
        else:
            pressure = 0.0  # 0.0 indicates not calculated. To be changed.

        forces_out = (
            field_forces
            + bond_forces
            + angle_forces
            + dihedral_forces
            + reconstructed_forces
        )
        if charges_flag:
            forces_out += elec_forces

    # initialize PLUMED and get initial forces and bias
    if args.plumed:
        plumed = PlumedBias(
            config, 
            args.plumed, 
            args.plumed_outfile, 
            comm=comm,
            verbose=args.verbose
        )

        current_forces = (bond_forces + angle_forces + dihedral_forces
                        + field_forces + reconstructed_forces)
        if charges_flag:
            current_forces += elec_forces

        # pass info and check if PLUMED also needs the energy
        needs_energy = plumed.prepare(
            0,
            current_forces,
            positions,
            indices,
            config,
            (
                np.zeros_like(indices, dtype=np.double) 
                if not charges_flag else charges
            )
        )

        # PLUMED likes the energy per rank
        poteng = (field_energy + bond_energy + angle_energy
                 + dihedral_energy + field_q_energy) / size

        # get the forces and bias from PLUMED
        plumed_forces, plumed_bias = plumed.calc(current_forces, poteng)

    # log after PLUMED forces and bias have been calculated
    if config.n_print > 0:
        if args.plumed:
            forces_out += plumed_forces

        store_data(
            out_dataset,
            step,
            frame,
            indices,
            positions,
            velocities,
            forces_out,
            config.box_size,
            temperature,
            pressure,
            kinetic_energy,
            bond_energy,
            angle_energy,
            dihedral_energy,
            field_energy,
            field_q_energy,
            plumed_bias,
            config.time_step,
            config,
            velocity_out=args.velocity_output,
            force_out=args.force_output,
            charge_out=charges_flag,
            plumed_out=True if args.plumed else False,
            dump_per_particle=args.dump_per_particle,
            comm=comm,
        )

    if rank == 0:
        loop_start_time = datetime.datetime.now()
        last_step_time = datetime.datetime.now()

    # MD loop
    for step in range(1, config.n_steps + 1):
        current_step_time = datetime.datetime.now()

        if step == 1 and args.verbose > 1:
            Logger.rank0.log(logging.INFO, f"MD step = {step:10d}")
        else:
            log_step = False
            if config.n_steps < 1000:
                log_step = True
            elif (
                np.mod(step, config.n_steps // 1000) == 0
                or np.mod(step, config.n_print) == 0
            ):
                log_step = True
            if rank == 0 and log_step and args.verbose > 1:
                step_t = current_step_time - last_step_time
                tot_t = current_step_time - loop_start_time
                ns_sim = (step + 1) * config.respa_inner * config.time_step / 1000.0

                seconds_per_day = 24 * 60 * 60
                seconds_elapsed = tot_t.days * seconds_per_day
                seconds_elapsed += tot_t.seconds
                seconds_elapsed += 1e-6 * tot_t.microseconds
                minutes_elapsed = seconds_elapsed / 60
                hours_elapsed = minutes_elapsed / 60
                days_elapsed = hours_elapsed / 24

                ns_per_day = ns_sim / days_elapsed
                hours_per_ns = hours_elapsed / ns_sim
                steps_per_s = (step + 1) / seconds_elapsed
                info_str = (
                    f"MD step = {step:10d}   step time: "
                    f"{format_timedelta(step_t):22s}   Performance: "
                    f"{ns_per_day:.3f} ns/day   {hours_per_ns:.3f} hours/ns   "
                    f"{steps_per_s:.3f} steps/s"
                )
                Logger.rank0.log(logging.INFO, info_str)

        # Initial outer rRESPA velocity step
        velocities = integrate_velocity(
            velocities,
            field_forces / config.mass,
            config.respa_inner * config.time_step,
        )
        if args.plumed:
            velocities = integrate_velocity(
                velocities, plumed_forces / config.mass,
                config.respa_inner * config.time_step,
            )
        if charges_flag and (
            config.coulombtype == "PIC_Spectral"
            or config.coulombtype == "PIC_Spectral_GPE"
        ):
            velocities = integrate_velocity(
                velocities,
                elec_forces / config.mass,
                config.respa_inner * config.time_step,
            )
        if protein_flag:
            velocities = integrate_velocity(
                velocities,
                reconstructed_forces / config.mass,
                config.respa_inner * config.time_step,
            )

        # Inner rRESPA steps
        for inner in range(config.respa_inner):
            velocities = integrate_velocity(
                velocities,
                (bond_forces + angle_forces + dihedral_forces) / config.mass,
                config.time_step,
            )
            positions = integrate_position(positions, velocities, config.time_step)
            positions = np.mod(positions, config.box_size[None, :])

            # Update fast forces
            if molecules_flag:
                if not args.disable_bonds:
                    bond_energy_, bond_pr_ = compute_bond_forces(
                        bond_forces,
                        positions,
                        config.box_size,
                        bonds_2_atom1,
                        bonds_2_atom2,
                        bonds_2_equilibrium,
                        bonds_2_stength,
                    )
                if not args.disable_angle_bonds:
                    angle_energy_, angle_pr_ = compute_angle_forces(
                        angle_forces,
                        positions,
                        config.box_size,
                        bonds_3_atom1,
                        bonds_3_atom2,
                        bonds_3_atom3,
                        bonds_3_equilibrium,
                        bonds_3_stength,
                    )
                if not args.disable_dihedrals:
                    if inner == config.respa_inner - 1 and not args.disable_dipole:
                        dipole_flag = 1
                    else:
                        dipole_flag = 0
                    dihedral_energy_ = compute_dihedral_forces(
                        dihedral_forces,
                        positions,
                        dipole_positions,
                        transfer_matrices,
                        config.box_size,
                        bonds_4_atom1,
                        bonds_4_atom2,
                        bonds_4_atom3,
                        bonds_4_atom4,
                        bonds_4_coeff,
                        bonds_4_type,
                        bonds_4_last,
                        dipole_flag,
                    )

            velocities = integrate_velocity(
                velocities,
                (bond_forces + angle_forces + dihedral_forces) / config.mass,
                config.time_step,
            )

        # Barostat
        if config.barostat:
            if config.barostat.lower() == "isotropic":
                pm_objs, change = isotropic(
                    pmesh,
                    pm_objs,
                    phi,
                    phi_q,
                    psi,
                    hamiltonian,
                    positions,
                    velocities,
                    config,
                    phi_fourier,
                    phi_laplacian,
                    phi_transfer,
                    bond_pr_,
                    angle_pr_,
                    step,
                    prng,
                    comm=comm,
                )

            elif config.barostat.lower() == "semiisotropic":
                pm_objs, change = semiisotropic(
                    pmesh,
                    pm_objs,
                    phi,
                    phi_q,
                    psi,
                    hamiltonian,
                    positions,
                    velocities,
                    config,
                    phi_fourier,
                    phi_laplacian,
                    phi_transfer,
                    bond_pr_,
                    angle_pr_,
                    step,
                    prng,
                    comm=comm,
                )
            pm, field_list, elec_common_list, coulomb_list = pm_objs
            (
                phi,
                phi_fourier,
                force_on_grid,
                v_ext_fourier,
                v_ext,
                phi_transfer,
                phi_laplacian,
            ) = field_list

            phi_q, phi_q_fourier, psi, elec_field = elec_common_list

            if len(coulomb_list) == 2:
                elec_field_fourier, psi_fourier = coulomb_list

            elif len(coulomb_list) == 12:
                (
                    phi_eps,
                    phi_eps_fourier,
                    phi_eta,
                    phi_eta_fourier,
                    phi_pol,
                    phi_pol_prev,
                    elec_dot,
                    elec_field_contrib,
                    psi,
                    Vbar_elec,
                    Vbar_elec_fourier,
                    force_mesh_elec,
                    force_mesh_elec_fourier,
                ) = coulomb_list

        # Only compute and keep the molecular bond energy from the last rRESPA
        # inner step
        if molecules_flag:
            if not args.disable_bonds:
                bond_energy = comm.allreduce(bond_energy_, MPI.SUM)
            if not args.disable_angle_bonds:
                angle_energy = comm.allreduce(angle_energy_, MPI.SUM)
            if not args.disable_dihedrals:
                dihedral_energy = comm.allreduce(dihedral_energy_, MPI.SUM)

        # Update slow forces
        if not args.disable_field:
            layouts = [
                pm.decompose(positions[types == t])
                for t in range(config.n_types)  # noqa: E501
            ]
            update_field(
                phi,
                phi_laplacian,
                phi_transfer,
                layouts,
                force_on_grid,
                hamiltonian,
                pm,
                positions,
                types,
                config,
                v_ext,
                phi_fourier,
                v_ext_fourier,
                config.m,
            )
            compute_field_force(
                layouts,
                positions,
                force_on_grid,
                field_forces,
                types,
                config.n_types,
            )

            if charges_flag:
                layout_q = pm.decompose(positions)
                if config.coulombtype == "PIC_Spectral_GPE":  # dielectric_flag and
                    Vbar_elec, phi_eps, elec_dot = update_field_force_q_GPE(
                        conv_fun,
                        phi,
                        types,
                        charges,
                        phi_q,
                        phi_q_fourier,
                        phi_eps,
                        phi_eps_fourier,
                        phi_eta,
                        phi_eta_fourier,
                        phi_pol_prev,
                        phi_pol,
                        elec_field,
                        elec_forces,
                        elec_field_contrib,
                        psi,
                        Vbar_elec,
                        Vbar_elec_fourier,
                        force_mesh_elec,
                        force_mesh_elec_fourier,
                        hamiltonian,
                        layout_q,
                        layouts,
                        pm,
                        positions,
                        config,
                        comm=comm,
                    )

                    field_q_energy = compute_field_energy_q_GPE(
                        config, phi_eps, field_q_energy, elec_dot, comm=comm
                    )

                if config.coulombtype == "PIC_Spectral":
                    update_field_force_q(
                        charges,
                        phi_q,
                        phi_q_fourier,
                        psi,
                        psi_fourier,
                        elec_field_fourier,
                        elec_field,
                        elec_forces,
                        layout_q,
                        hamiltonian,
                        pm,
                        positions,
                        config,
                    )

            if protein_flag and not args.disable_dipole:
                dipole_positions = np.reshape(dipole_positions, (4 * n_tors, 3))
                dipole_forces = np.reshape(dipole_forces, (4 * n_tors, 3))

                layout_dipoles = pm.decompose(dipole_positions)
                update_field_force_q(
                    dipole_charges,
                    phi_dipoles,
                    phi_dipoles_fourier,
                    psi_dipoles,
                    psi_dipoles_fourier,
                    dipoles_field_fourier,
                    dipoles_field,
                    dipole_forces,
                    layout_dipoles,
                    hamiltonian,
                    pm,
                    dipole_positions,
                    config,
                )

                dipole_positions = np.reshape(dipole_positions, (n_tors, 4, 3))
                dipole_forces = np.reshape(dipole_forces, (n_tors, 4, 3))

                dipole_positions = np.asfortranarray(dipole_positions)
                dipole_forces_redistribution(
                    reconstructed_forces,
                    dipole_forces,
                    transfer_matrices,
                    bonds_4_atom1,
                    bonds_4_atom2,
                    bonds_4_atom3,
                    bonds_4_atom4,
                    bonds_4_type,
                    bonds_4_last,
                )

        if args.plumed:
            current_forces = (bond_forces + angle_forces + dihedral_forces
                            + field_forces + reconstructed_forces)
            if charges_flag:
                current_forces += elec_forces

            # pass info and check if PLUMED also needs the energy
            needs_energy = plumed.prepare(
                step,
                current_forces,
                positions,
                indices,
                config,
                (
                    np.zeros_like(indices, dtype=np.double) 
                    if not charges_flag else charges
                )
            )

            if needs_energy:
                if not args.disable_field:
                    (
                        field_energy, kinetic_energy,
                    ) = compute_field_and_kinetic_energy(
                        phi, velocities, hamiltonian, positions, types, v_ext,
                        config, layouts, comm=comm,
                    )

                    if charges_flag and config.coulombtype == "PIC_Spectral":
                        field_q_energy = compute_field_energy_q(
                            config, phi_q_fourier, elec_energy_field,
                            field_q_energy, comm=comm,
                        )
                else:
                    field_energy = 0.0
                    field_q_energy = 0.0

                # PLUMED likes the energy per rank
                poteng = (field_energy + bond_energy + angle_energy
                         + dihedral_energy + field_q_energy) / size
            else:
                poteng = 0.0

            # get the forces and bias from PLUMED
            plumed_forces, plumed_bias = plumed.calc(current_forces, poteng)

        # Second rRESPA velocity step
        velocities = integrate_velocity(
            velocities,
            field_forces / config.mass,
            config.respa_inner * config.time_step,
        )

        if args.plumed:
            velocities = integrate_velocity(
                velocities, plumed_forces / config.mass,
                config.respa_inner * config.time_step,
            )
        if charges_flag and (
            config.coulombtype == "PIC_Spectral"
            or config.coulombtype == "PIC_Spectral_GPE"
        ):
            velocities = integrate_velocity(
                velocities,
                elec_forces / config.mass,
                config.respa_inner * config.time_step,
            )
        if protein_flag and not args.disable_dipole:
            velocities = integrate_velocity(
                velocities,
                reconstructed_forces / config.mass,
                config.respa_inner * config.time_step,
            )

        if config.domain_decomposition:
            if np.mod(step, config.domain_decomposition) == 0:
                positions = np.ascontiguousarray(positions)
                bond_forces = np.ascontiguousarray(bond_forces)
                angle_forces = np.ascontiguousarray(angle_forces)
                dihedral_forces = np.ascontiguousarray(dihedral_forces)
                if args.plumed:
                    plumed_forces = np.ascontiguousarray(plumed_forces)

                (
                    positions,
                    velocities,
                    indices,
                    bond_forces,
                    angle_forces,
                    dihedral_forces,
                    reconstructed_forces,
                    field_forces,
                    names,
                    types,
                    *optional,
                ) = domain_decomposition(
                    positions,
                    pm,
                    *tuple(args_in),
                    molecules=molecules if molecules_flag else None,
                    bonds=bonds if molecules_flag else None,
                    verbose=args.verbose,
                    comm=comm,
                )

                if charges_flag:
                    charges = optional.pop(0)
                    elec_forces = optional.pop(0)
                if dielectric_flag:
                    dielectric_sorted = optional.pop(0)
                if args.plumed:
                    plumed_forces = optional.pop(0)
                if molecules_flag:
                    bonds = optional.pop(0)
                    molecules = optional.pop(0)

                # rebuild args_in to point to correct arrays
                args_in = [
                    velocities,
                    indices,
                    bond_forces,
                    angle_forces,
                    dihedral_forces,
                    reconstructed_forces,
                    field_forces,
                    names,
                    types,
                ]

                if charges_flag:
                    args_in.append(charges)
                    args_in.append(elec_forces)
                if dielectric_flag:
                    args_in.append(dielectric_sorted)
                if args.plumed:
                    args_in.append(plumed_forces)

                positions = np.asfortranarray(positions)
                bond_forces = np.asfortranarray(bond_forces)
                angle_forces = np.asfortranarray(angle_forces)
                dihedral_forces = np.asfortranarray(dihedral_forces)

                layouts = [
                    pm.decompose(positions[types == t])
                    for t in range(config.n_types)  # noqa: E501
                ]
                if molecules_flag:
                    bonds_prep = prepare_bonds(
                        molecules, names, bonds, indices, config, topol
                    )
                    (
                        # two-particle bonds
                        bonds_2_atom1,
                        bonds_2_atom2,
                        bonds_2_equilibrium,
                        bonds_2_stength,
                        # three-particle bonds
                        bonds_3_atom1,
                        bonds_3_atom2,
                        bonds_3_atom3,
                        bonds_3_equilibrium,
                        bonds_3_stength,
                        # four-particle bonds
                        bonds_4_atom1,
                        bonds_4_atom2,
                        bonds_4_atom3,
                        bonds_4_atom4,
                        bonds_4_coeff,
                        bonds_4_type,
                        bonds_4_last,
                    ) = bonds_prep

                    bonds_4_coeff = np.asfortranarray(bonds_4_coeff)

                    # Reinitialize dipoles so each rank has the right amount
                    if protein_flag and not args.disable_dipole:
                        # Each rank will have different n_tors, we don't need
                        # to domain decompose dipoles
                        n_tors = len(bonds_4_atom1)
                        dipole_positions = np.zeros((n_tors, 4, 3), dtype=dtype)

                        # 4 because we need to take into account the last angle
                        # in the molecule
                        dipole_charges = np.array(
                            [
                                2 * [0.25, -0.25]
                                if (bonds_4_type[i], bonds_4_last[i]) == (1, 1)
                                else [0.25, -0.25, 0.0, 0.0]
                                if bonds_4_type[i] == 1
                                else 2 * [0.0, 0.0]
                                for i in range(n_tors)
                            ],
                            dtype=dtype,
                        ).flatten()
                        dipole_forces = np.zeros_like(dipole_positions)
                        transfer_matrices = np.zeros(
                            shape=(n_tors, 6, 3, 3), dtype=dtype
                        )
                        dipole_positions = np.asfortranarray(dipole_positions)
                        transfer_matrices = np.asfortranarray(
                            transfer_matrices
                        )  # noqa: E501

        for t in range(config.n_types):
            if args.verbose > 2:
                exchange_cost = layouts[t].get_exchange_cost()
                Logger.all_ranks.log(
                    logging.INFO,
                    (
                        f"(GHOSTS: Total number of particles of type "
                        f"{config.type_to_name_map[t]} to be "
                        f"exchanged = {exchange_cost[rank]}"
                    ),
                )

        # Thermostat
        if config.target_temperature and np.mod(step, config.n_b) == 0:
            csvr_thermostat(velocities, names, config, prng, comm=comm)

        # Remove total linear momentum
        if config.cancel_com_momentum:
            if np.mod(step, config.cancel_com_momentum) == 0:
                velocities = cancel_com_momentum(velocities, config, comm=comm)

        # Print trajectory
        if config.n_print > 0:
            if np.mod(step, config.n_print) == 0:
                frame = step // config.n_print
                if not args.disable_field:
                    layouts = [
                        pm.decompose(positions[types == t])
                        for t in range(config.n_types)
                    ]
                    update_field(
                        phi,
                        phi_laplacian,
                        phi_transfer,
                        layouts,
                        force_on_grid,
                        hamiltonian,
                        pm,
                        positions,
                        types,
                        config,
                        v_ext,
                        phi_fourier,
                        v_ext_fourier,
                        config.m,
                        compute_potential=True,
                    )
                    (
                        field_energy,
                        kinetic_energy,
                        field_q_energy,
                    ) = compute_field_and_kinetic_energy(
                        phi,
                        phi_q,
                        psi,
                        velocities,
                        hamiltonian,
                        positions,
                        types,
                        v_ext,
                        config,
                        layouts,
                        comm=comm,
                    )

                    if charges_flag:
                        if config.coulombtype == "PIC_Spectral_GPE":
                            field_q_energy = compute_field_energy_q_GPE(
                                config, phi_eps, field_q_energy, elec_dot, comm=comm
                            )
                else:
                    kinetic_energy = comm.allreduce(
                        0.5 * config.mass * np.sum(velocities**2)
                    )
                temperature = (
                    (2 / 3)
                    * kinetic_energy
                    / (config.gas_constant * config.n_particles)
                )
                if args.disable_field:
                    field_energy = 0.0

                if config.pressure or config.barostat:
                    pressure = comp_pressure(
                        phi,
                        phi_q,
                        psi,
                        hamiltonian,
                        velocities,
                        config,
                        phi_fourier,
                        phi_laplacian,
                        phi_transfer,
                        positions,
                        bond_pr_,
                        angle_pr_,
                        comm=comm,
                    )
                else:
                    pressure = 0.0  # 0.0 indicates not calculated. To be changed.

                forces_out = (
                    field_forces
                    + bond_forces
                    + angle_forces
                    + dihedral_forces
                    + reconstructed_forces
                )
                if charges_flag:
                    forces_out += elec_forces
                if args.plumed:
                    forces_out += plumed_forces
                store_data(
                    out_dataset,
                    step,
                    frame,
                    indices,
                    positions,
                    velocities,
                    forces_out,
                    config.box_size,
                    temperature,
                    pressure,
                    kinetic_energy,
                    bond_energy,
                    angle_energy,
                    dihedral_energy,
                    field_energy,
                    field_q_energy,
                    plumed_bias,
                    config.respa_inner * config.time_step,
                    config,
                    velocity_out=args.velocity_output,
                    force_out=args.force_output,
                    charge_out=charges_flag,
                    plumed_out=True if args.plumed else False,
                    dump_per_particle=args.dump_per_particle,
                    comm=comm,
                )
                if np.mod(step, config.n_print * config.n_flush) == 0:
                    out_dataset.flush()
        last_step_time = current_step_time

    # End simulation
    if rank == 0:
        end_time = datetime.datetime.now()
        sim_time = end_time - start_time
        setup_time = loop_start_time - start_time
        loop_time = end_time - loop_start_time
        Logger.rank0.log(
            logging.INFO,
            (
                f"Elapsed time: {format_timedelta(sim_time)}   "
                f"Setup time: {format_timedelta(setup_time)}   "
                f"MD loop time: {format_timedelta(loop_time)}"
            ),
        )

    if config.n_print > 0 and np.mod(config.n_steps, config.n_print) != 0:
        if not args.disable_field:
            update_field(
                phi,
                phi_laplacian,
                phi_transfer,
                layouts,
                force_on_grid,
                hamiltonian,
                pm,
                positions,
                types,
                config,
                v_ext,
                phi_fourier,
                v_ext_fourier,
                config.m,
                compute_potential=True,
            )
            (
                field_energy,
                kinetic_energy,
                field_q_energy,
            ) = compute_field_and_kinetic_energy(
                phi,
                phi_q,
                psi,
                velocities,
                hamiltonian,
                positions,
                types,
                v_ext,
                config,
                layouts,
                comm=comm,
            )

            if charges_flag:
                layout_q = pm.decompose(positions)
                if config.coulombtype == "PIC_Spectral_GPE":
                    Vbar_elec, phi_eps, elec_dot = update_field_force_q_GPE(
                        conv_fun,
                        phi,
                        types,
                        charges,
                        phi_q,
                        phi_q_fourier,
                        phi_eps,
                        phi_eps_fourier,
                        phi_eta,
                        phi_eta_fourier,
                        phi_pol_prev,
                        phi_pol,
                        elec_field,
                        elec_forces,
                        elec_field_contrib,
                        psi,
                        Vbar_elec,
                        Vbar_elec_fourier,
                        force_mesh_elec,
                        force_mesh_elec_fourier,
                        hamiltonian,
                        layout_q,
                        layouts,
                        pm,
                        positions,
                        config,
                        comm=comm,
                    )

                    field_q_energy = compute_field_energy_q_GPE(
                        config, phi_eps, field_q_energy, elec_dot, comm=comm
                    )

                if config.coulombtype == "PIC_Spectral":
                    update_field_force_q(
                        charges,
                        phi_q,
                        phi_q_fourier,
                        psi,
                        psi_fourier,
                        elec_field_fourier,
                        elec_field,
                        elec_forces,
                        layout_q,
                        hamiltonian,
                        pm,
                        positions,
                        config,
                    )

        else:
            kinetic_energy = comm.allreduce(0.5 * config.mass * np.sum(velocities**2))
        frame = (step + 1) // config.n_print
        temperature = (
            (2 / 3) * kinetic_energy / (config.gas_constant * config.n_particles)
        )
        if args.disable_field:
            field_energy = 0.0

        if config.pressure or config.barostat:
            pressure = comp_pressure(
                phi,
                phi_q,
                psi,
                hamiltonian,
                velocities,
                config,
                phi_fourier,
                phi_laplacian,
                phi_transfer,
                positions,
                bond_pr_,
                angle_pr_,
                Vbar_elec,
                comm=comm,
            )
        else:
            pressure = 0.0  # 0.0 indicates not calculated. To be changed.

        forces_out = (
            field_forces
            + bond_forces
            + angle_forces
            + dihedral_forces
            + reconstructed_forces
        )
        if charges_flag:
            forces_out += elec_forces
        if args.plumed:
            forces_out += plumed_forces
        store_data(
            out_dataset,
            step,
            frame,
            indices,
            positions,
            velocities,
            forces_out,
            config.box_size,
            temperature,
            pressure,
            kinetic_energy,
            bond_energy,
            angle_energy,
            dihedral_energy,
            field_energy,
            field_q_energy,
            plumed_bias,
            config.respa_inner * config.time_step,
            config,
            velocity_out=args.velocity_output,
            force_out=args.force_output,
            charge_out=charges_flag,
            plumed_out=True if args.plumed else False,
            dump_per_particle=args.dump_per_particle,
            comm=comm,
        )

    out_dataset.close_file()
    if args.plumed:
        plumed.finalize()
